#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
import logging
import os
import re
import shutil
import glob
import datetime
import numpy as np
import pandas as pd
import xarray as xr
import dask.dataframe as dd
import importlib.metadata
from typing import Union
from disdrodb.utils.logger import log_info, log_warning
from disdrodb.L0.metadata import read_metadata

logger = logging.getLogger(__name__)

####---------------------------------------------------------------------------.
#### Directory/Filepaths Defaults
def infer_institute_from_fpath(fpath: str) -> str:
    """Infer institue name from file path.

    Parameters
    ----------
    fpath : str
        Input file path.

    Returns
    -------
    str
        Name of the institute.
    """
    idx_start = fpath.rfind("DISDRODB")
    disdrodb_fpath = fpath[idx_start:]
    institute = disdrodb_fpath.split("/")[2]
    return institute


def infer_campaign_from_fpath(fpath: str) -> str:
    """Infer campaign name from file path.

    Parameters
    ----------
    fpath : str
        Input file path.

    Returns
    -------
    str
        Name of the campaign.
    """
    idx_start = fpath.rfind("DISDRODB")
    disdrodb_fpath = fpath[idx_start:]
    campaign = disdrodb_fpath.split("/")[3]
    return campaign


def infer_station_id_from_fpath(fpath: str) -> str:
    """
    Get the station ID from the path of the input raw data.

    Parameters
    ----------
    fpath : str
        Path of the raw file.

    Returns
    -------
    str
        Station ID
    """
    path_pattern = r"(\\|\/)"
    idx_start = fpath.rfind("DISDRODB")
    disdrodb_fpath = fpath[idx_start:]
    list_path_elements = re.split(path_pattern, disdrodb_fpath)
    station_id = list_path_elements[10]
    # Optional strip .yml if fpath point to YAML file
    station_id.strip(".yml")
    return station_id


def get_campaign_name(base_dir: str) -> str:
    """Return the campaign name from 'raw_dir' or 'processed_dir' paths.

    Parameters
    ----------
    base_dir : str
        Path 'raw_dir' or 'processed_dir' directory.

    Returns
    -------
    str
        Name of the campaign.
    """

    base_dir = parse_fpath(base_dir)
    campaign_name = os.path.basename(base_dir).upper()
    return campaign_name


def get_data_source(base_dir: str) -> str:
    """Retrieves the data source from 'raw_dir' or processed_dir' paths

    Parameters
    ----------
    base_dir : str
        Input paths

    Returns
    -------
    str
        Name of the data source
    """

    base_dir = parse_fpath(base_dir)
    institute_name = os.path.basename(os.path.dirname(base_dir)).upper()
    return institute_name


def get_dataset_min_max_time(ds: xr.Dataset):
    """Retrieves dataset starting and ending time.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset

    Returns
    -------
    tuple
        (starting_time, ending_time)

    """

    starting_time = ds["time"].values[0]
    ending_time = ds["time"].values[-1]
    return (starting_time, ending_time)


# TODO: get_dataframe_min_max_time


def get_L0A_dir(processed_dir: str, station_id: str) -> str:
    """Get L0A directory

    Parameters
    ----------
    processed_dir : str
        Path of the processed directory
    station_id : int
        ID of the station

    Returns
    -------
    str
        L0A directory path.
    """
    dir_path = os.path.join(processed_dir, "L0A", station_id)
    return dir_path


def get_L0A_fname(campaign_name: str, station_id: str, suffix: str = "") -> str:
    """build L0A file name.

    Parameters
    ----------
    campaign_name : str
        Name of the campaign.
    station_id : int
        ID of the station
    suffix : int, optional
        suffix, by default ""

    Returns
    -------
    str
        L0A file name.
    """
    if suffix != "":
        suffix = "_" + suffix
    fname = campaign_name + "_s" + station_id + suffix + ".parquet"
    return fname


# TODO: and refactor L0_processing --> remove suffix
#
# def get_L0A_fname(df, processed_dir, station_id: str) -> str:
#     """Define L0A file name.

#     Parameters
#     ----------
#     ds : pd.DataFrame
#         L0A DataFrame
#     processed_dir : str
#         Path of the processed directory
#     station_id : int
#         ID of the station

#     Returns
#     -------
#     str
#         L0B file name.
#     """
#     starting_time, ending_time = get_dataframe_min_max_time(ds)
#     starting_time = pd.to_datetime(starting_time).strftime("%Y%m%d%H%M%S")
#     ending_time = pd.to_datetime(ending_time).strftime("%Y%m%d%H%M%S")
#     # production_time = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
#     campaign_name = get_campaign_name(processed_dir).replace(".", "-")
#     institute_name = get_data_source(processed_dir).replace(".", "-")
#     metadata_dict = read_metadata(processed_dir, station_id)
#     sensor_name = metadata_dict.get("sensor_name").replace("_", "-")
#     version = importlib.metadata.version("disdrodb").replace(".", "-")
#     if version == "-VERSION-PLACEHOLDER-":
#         version = "dev"
#     fname = f"DISDRODB.L0A.Raw.{institute_name}.{campaign_name}.{station_id}.{sensor_name}.s{starting_time}.e{ending_time}.{version}.parquet"
#     return fname


def get_L0A_fpath(processed_dir: str, station_id: str, suffix: str = "") -> str:
    """build L0A file path.

    Parameters
    ----------
    campaign_name : str
        Name of the campaign.
    station_id : int
        ID of the station
    suffix : int, optional
        suffix, by default ""

    Returns
    -------
    str
        L0A file path.
    """
    campaign_name = get_campaign_name(processed_dir)
    fname = get_L0A_fname(campaign_name, station_id, suffix=suffix)
    dir_path = get_L0A_dir(processed_dir, station_id)
    fpath = os.path.join(dir_path, fname)
    return fpath


def get_L0B_dir(processed_dir: str, station_id: str) -> str:
    """Build L0B directory

    Parameters
    ----------
    processed_dir : str
        Path of the processed directory
    station_id : int
        ID of the station

    Returns
    -------
    str
        Path of the L0B directory
    """
    dir_path = os.path.join(processed_dir, "L0B", station_id)
    return dir_path


def get_L0B_fname(ds, processed_dir, station_id: str) -> str:
    """Define L0B file name.

    Parameters
    ----------
    ds : xr.Dataset
        L0B xarray Dataset
    processed_dir : str
        Path of the processed directory
    station_id : int
        ID of the station

    Returns
    -------
    str
        L0B file name.
    """
    starting_time, ending_time = get_dataset_min_max_time(ds)
    starting_time = pd.to_datetime(starting_time).strftime("%Y%m%d%H%M%S")
    ending_time = pd.to_datetime(ending_time).strftime("%Y%m%d%H%M%S")
    # production_time = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    # institute_name = get_data_source(processed_dir).replace(".", "-") # TODO: data_source
    campaign_name = get_campaign_name(processed_dir).replace(".", "-")
    metadata_dict = read_metadata(processed_dir, station_id)
    sensor_name = metadata_dict.get("sensor_name").replace("_", "-")
    version = importlib.metadata.version("disdrodb").replace(".", "-")
    if version == "-VERSION-PLACEHOLDER-":
        version = "dev"
    fname = f"DISDRODB.L0B.Raw.{campaign_name}.{station_id}.{sensor_name}.s{starting_time}.e{ending_time}.{version}.nc"
    return fname


def get_L0B_fpath(ds, processed_dir: str, station_id: str) -> str:
    """Define L0B file path.

    Parameters
    ----------
    ds : xr.Dataset
        L0B xarray Dataset
    processed_dir : str
        Path of the processed directory
    station_id : int
        ID of the station

    Returns
    -------
    str
        L0B file path.
    """
    dir_path = get_L0B_dir(processed_dir, station_id)
    fname = get_L0B_fname(ds, processed_dir, station_id)
    fpath = os.path.join(dir_path, fname)
    return fpath


####--------------------------------------------------------------------------.
#### Directory/File Creation/Deletion


def _create_directory(path: str) -> None:
    if not isinstance(path, str):
        raise TypeError("'path' must be a strig.")
    try:
        os.makedirs(path)
        logger.debug(f"Created directory {path}.")
    except FileExistsError:
        logger.debug(f"Directory {path} already exists.")
        pass
    except (Exception) as e:
        dir_name = os.path.basename(path)
        msg = f"Can not create folder {dir_name} inside <path>. Error: {e}"
        logger.exception(msg)
        raise FileNotFoundError(msg)


def _remove_if_exists(fpath: str, force: bool = False) -> None:
    if os.path.exists(fpath):
        if not force:
            msg = f"--force is False and a file already exists at:{fpath}"
            logger.error(msg)
            raise ValueError(msg)
        try:
            os.remove(fpath)
        except IsADirectoryError:
            try:
                os.rmdir(fpath)
            except OSError:
                try:
                    # shutil.rmtree(fpath.rpartition('.')[0])
                    for f in glob.glob(fpath + "/*"):
                        try:
                            os.remove(f)
                        except OSError as e:
                            msg = f"Can not delete file {f}, error: {e.strerror}"
                            logger.exception(msg)
                    os.rmdir(fpath)
                except:
                    msg = f"Something wrong with: {fpath}"
                    logger.error(msg)
                    raise ValueError(msg)
        logger.info(f"Deleted folder {fpath}")


####--------------------------------------------------------------------------.
#### Directory checks
def parse_fpath(fpath: str) -> str:
    """Ensure fpath does not end with /.

    Parameters
    ----------
    fpath : str
        Input file path

    Returns
    -------
    str
        Output file path

    Raises
    ------
    TypeError
        Error il file path not compliant
    """

    if not isinstance(fpath, str):
        raise TypeError("'parse_fpath' expects a directory/filepath string.")
    if fpath[-1] == "/":
        print("{} should not end with /.".format(fpath))
        fpath = fpath[:-1]
    return fpath


####--------------------------------------------------------------------------.
#### L0 processing directory checks
def check_raw_dir(raw_dir: str, verbose: bool = False) -> None:
    """Check validity of raw_dir.

    Steps:
    1. Check that 'raw_dir' is a valid directory path
    2. Check that 'raw_dir' follows the expect directory structure
    3. Check that each station_id directory contains data
    4. Check that for each station_id the mandatory metadata are specified.



    Parameters
    ----------
    raw_dir : str
        Input raw directory
    verbose : bool, optional
        Wheter to verbose the processing.
        The default is False.

    Raises
    ------
    TypeError
        Error if not complient.
    """

    from disdrodb.L0.metadata import create_metadata
    from disdrodb.L0.metadata import check_metadata_compliance
    from disdrodb.L0.issue import create_issue_yml
    from disdrodb.L0.issue import check_issue_compliance

    # -------------------------------------------------------------------------.
    # Check input argument
    if not isinstance(raw_dir, str):
        raise TypeError("Provide 'raw_dir' as a string'.")
    if not os.path.exists(raw_dir):
        raise ValueError("'raw_dir' {} directory does not exist.".format(raw_dir))
    if not os.path.isdir(raw_dir):
        raise ValueError("'raw_dir' {} is not a directory.".format(raw_dir))
    # -------------------------------------------------------------------------.
    #### Check there is /data subfolders
    list_subfolders = os.listdir(raw_dir)
    if len(list_subfolders) == 0:
        raise ValueError("There are not subfolders in {}".format(raw_dir))
    if "data" not in list_subfolders:
        raise ValueError(
            "'raw_dir' {} should have the /data subfolder.".format(raw_dir)
        )

    # -------------------------------------------------------------------------.
    #### Check there are subfolders corresponding to station to process
    raw_data_dir = os.path.join(raw_dir, "data")
    list_data_station_id = os.listdir(raw_data_dir)
    if len(list_data_station_id) == 0:
        raise ValueError("No station directories within {}".format(raw_data_dir))

    # -------------------------------------------------------------------------.
    #### Check there are data files in each list_data_station_id
    list_raw_data_station_dir = [
        os.path.join(raw_data_dir, station_id) for station_id in list_data_station_id
    ]
    list_nfiles_per_station = [
        len(glob.glob(os.path.join(path, "*"))) for path in list_raw_data_station_dir
    ]
    idx_0_files = np.where(np.array(list_nfiles_per_station) == 0)[0]
    if len(idx_0_files) > 0:
        empty_station_dir = [list_raw_data_station_dir[idx] for idx in idx_0_files]
        raise ValueError(
            "The following data directories are empty: {}".format(empty_station_dir)
        )

    # -------------------------------------------------------------------------.
    #### Check there is /metadata subfolders
    metadata_dir = os.path.join(raw_dir, "metadata")
    if "metadata" not in list_subfolders:
        # - Create metadata directory
        _create_directory(metadata_dir)
        # - Create default metadata yml file for each station (since the folder didn't existed)
        list_metadata_fpath = [
            os.path.join(metadata_dir, station_id + ".yml")
            for station_id in list_data_station_id
        ]
        _ = [create_metadata(fpath) for fpath in list_metadata_fpath]
        msg = "'raw_dir' {} should have the /metadata subfolder. ".format(raw_dir)
        msg1 = "It has been now created with also empty metadata files to be filled for each station."
        raise ValueError(msg + msg1)

    # -------------------------------------------------------------------------.
    #### Check there are metadata file for each station_id in /metadata
    list_metadata_fpath = glob.glob(os.path.join(metadata_dir, "*.yml"))
    list_metadata_fname = [os.path.basename(fpath) for fpath in list_metadata_fpath]
    list_metadata_station_id = [fname[:-4] for fname in list_metadata_fname]

    # - Check there is metadata for each station
    missing_data_station_idx = np.where(
        np.isin(list_data_station_id, list_metadata_station_id, invert=True)
    )[0]
    # - If missing, create the defaults files and raise an error
    if len(missing_data_station_idx) > 0:
        list_missing_station_id = [
            list_data_station_id[idx] for idx in missing_data_station_idx
        ]
        list_missing_metadata_fpath = [
            os.path.join(metadata_dir, station_id + ".yml")
            for station_id in list_missing_station_id
        ]
        _ = [create_metadata(fpath) for fpath in list_missing_metadata_fpath]
        msg = "The metadata files for the following station_id were missing: {}".format(
            list_missing_station_id
        )
        raise ValueError(msg + " Now have been created to be filled.")

    # - Check not excess metadata compared to present stations
    excess_metadata_station_idx = np.where(
        np.isin(list_metadata_station_id, list_data_station_id, invert=True)
    )[0]
    if len(excess_metadata_station_idx) > 0:
        list_excess_station_id = [
            list_metadata_station_id[idx] for idx in excess_metadata_station_idx
        ]
        print(
            "There are the following metadata files without corresponding data: {}".format(
                list_excess_station_id
            )
        )

    # -------------------------------------------------------------------------.
    #### Check metadata compliance
    _ = [check_metadata_compliance(fpath) for fpath in list_metadata_fpath]
    # TODO: MISSING IMPLEMENTATION OF check_metadata_compliance
    # -------------------------------------------------------------------------.
    #### Check there is /issue subfolder
    issue_dir = os.path.join(raw_dir, "issue")
    if "issue" not in list_subfolders:
        # - Create issue directory
        _create_directory(issue_dir)
        # - Create issue yml file for each station (since the folder didn't existed)
        list_issue_fpath = [
            os.path.join(issue_dir, station_id + ".yml")
            for station_id in list_data_station_id
        ]
        _ = [create_issue_yml(fpath) for fpath in list_issue_fpath]
        msg = "The /issue subfolder has been now created to document and then remove timesteps with problematic data."
        logger.info(msg)
    # -------------------------------------------------------------------------.
    #### Check there are issue file for each station_id in /issue
    list_issue_fpath = glob.glob(os.path.join(issue_dir, "*.yml"))
    list_issue_fname = [os.path.basename(fpath) for fpath in list_issue_fpath]
    list_issue_station_id = [fname[:-4] for fname in list_issue_fname]

    # - Check there is issue for each station
    missing_data_station_idx = np.where(
        np.isin(list_data_station_id, list_issue_station_id, invert=True)
    )[0]
    # - If missing, create the defaults files and raise an error
    if len(missing_data_station_idx) > 0:
        list_missing_station_id = [
            list_data_station_id[idx] for idx in missing_data_station_idx
        ]
        list_missing_issue_fpath = [
            os.path.join(issue_dir, station_id + ".yml")
            for station_id in list_missing_station_id
        ]
        _ = [create_issue_yml(fpath) for fpath in list_missing_issue_fpath]
        msg = "The issue files for the following station_id were missing: {}".format(
            list_missing_station_id
        )

    # - Check not excess issue compared to present stations
    excess_issue_station_idx = np.where(
        np.isin(list_issue_station_id, list_data_station_id, invert=True)
    )[0]
    if len(excess_issue_station_idx) > 0:
        list_excess_station_id = [
            list_issue_station_id[idx] for idx in excess_issue_station_idx
        ]
        msg = f"There are the following issue files without corresponding data: {list_excess_station_id}"
        log_warning(logger, msg, verbose)

    # -------------------------------------------------------------------------.
    #### Check issue compliance
    _ = [check_issue_compliance(fpath) for fpath in list_issue_fpath]
    # TODO: MISSING IMPLEMENTATION OF check_issue_compliance
    # -------------------------------------------------------------------------.


def check_processed_dir(processed_dir: str, force: bool = False) -> None:
    """Check that 'processed_dir' is a valid directory path.

    Parameters
    ----------
    processed_dir : str
        Path of the processed directory
    force : bool, optional
        If True, overwrite existing data into processed directory.
        If False, raise an error if there are already data into processed directory.

    Raises
    ------
    TypeError
        Error if path pattern not respected or can not be created.


    """

    if not isinstance(processed_dir, str):
        raise TypeError("Provide 'processed_dir' as a string'.")
    # ------------------------------
    # Check processed_dir has "DISDRODB/Processed" to avoid deleting precious stuffs
    if (
        processed_dir.find("DISDRODB/Processed") == -1
        and processed_dir.find("DISDRODB\\Processed") == -1
    ):
        msg = "Expecting 'processed_dir' to contain the pattern */DISDRODB/Processed/*. or *\DISDRODB\Processed\*."
        logger.error(msg)
        raise ValueError(msg)

    # Check processed_dir does not end with "DISDRODB/Processed"
    # - It must contain also the <campaign_name> directory
    if (
        processed_dir.endswith("Processed")
        or processed_dir.endswith("Processed/")
        or processed_dir.endswith("Processed\\")
    ):
        msg = "Expecting 'processed_dir' to contain the pattern */DISDRODB/Processed/<campaign_name>."
        logger.error(msg)
        raise ValueError(msg)

    # ------------------------------
    # If forcing overwriting
    if force:
        # Check processed_dir is a directory before removing the content
        if os.path.exists(processed_dir):
            if not os.path.isdir(processed_dir):
                msg = f"'processed_dir' {processed_dir} already exist but is not a directory."
                logger.error(msg)
                raise ValueError(msg)
            # Remove content of existing processed_dir
            # TODO: maybe remove also the campaign name directory, not only the content of it
            shutil.rmtree(processed_dir)  # TODO: !! TOO DANGEROUS ??? !!!

    # ------------------------------
    # If avoiding overwriting
    if not force:
        if os.path.exists(processed_dir):
            msg = f"'processed_dir' {processed_dir} already exists and force=False."
            logger.error(msg)
            raise ValueError(msg)

    # ------------------------------
    # Recreate processed_dir
    if not os.path.exists(processed_dir):
        os.makedirs(processed_dir)
    else:
        msg = "Please report the BUG. This should not happen."
        logger.error(msg)
        raise ValueError(msg)


def check_campaign_name(raw_dir: str, processed_dir: str) -> str:
    """Check that 'raw_dir' and 'processed_dir' have same campaign_name.

    Parameters
    ----------
    raw_dir : str
        Path of the raw directory
    processed_dir : str
        Path of the processed directory

    Returns
    -------
    str
        Campaign name in capital letter

    Raises
    ------
    ValueError
        Error if both paths do not match.
    """

    upper_campaign_name = os.path.basename(raw_dir).upper()
    raw_campaign_name = os.path.basename(raw_dir)
    processed_campaign_name = os.path.basename(processed_dir)
    if raw_campaign_name != processed_campaign_name:
        msg = f"'raw_dir' and 'processed_dir' must ends with same <campaign_name> {upper_campaign_name}"
        logger.error(msg)
        raise ValueError(msg)

    if raw_campaign_name != upper_campaign_name:
        msg = f"'raw_dir' and 'processed_dir' must ends with UPPERCASE <campaign_name> {upper_campaign_name}"
        logger.error(msg)
        raise ValueError(msg)

    return upper_campaign_name


def check_directories(raw_dir: str, processed_dir: str, force: bool = False) -> tuple:
    """Check that the specified directories respect the standards.

    Parameters
    ----------
    raw_dir : str
        Path of the raw directory
    processed_dir : str
        Path of the processed directory
    force : bool, optional
        If True, overwrite existing data into processed directory.
        If False, raise an error if there are already data into processed directory.

    Returns
    -------
    tuple
        raw directory and processed directory
    """

    raw_dir = parse_fpath(raw_dir)
    processed_dir = parse_fpath(processed_dir)
    check_raw_dir(raw_dir)
    check_processed_dir(processed_dir, force=force)
    check_campaign_name(raw_dir, processed_dir)
    return raw_dir, processed_dir


def check_metadata_dir(processed_path: str) -> None:
    """Check metadata folder

    Parameters
    ----------
    processed_path : str
        Path of the processed directory

    Raises
    ------
    FileNotFoundError
        Error metadat already existed or can not be created.
    """
    # Create metadata folder
    try:
        metadata_folder_path = os.path.join(processed_path, "metadata")
        os.makedirs(metadata_folder_path)
        logger.debug(f"Created {metadata_folder_path}.")
    except FileExistsError:
        logger.debug(f"{metadata_folder_path} already existed.")
        pass
    except (Exception) as e:
        msg = f"Folder metadata can not be created in {metadata_folder_path}>. \n The error is: {e}."
        logger.exception(msg)
        raise FileNotFoundError(msg)


def copy_metadata_from_raw_dir(raw_dir: str, processed_dir: str) -> None:
    """Copy yaml files in raw_dir/metadata into processed_dir/metadata

    Parameters
    ----------
    raw_dir : str
        Path of the raw directory
    processed_dir : str
        Path of the processed directory

    Raises
    ------
    ValueError
        Error if the copy fails.
    """

    # Get src and dst metadata directory
    raw_metadata_dir = os.path.join(raw_dir, "metadata")
    processed_metadata_dir = os.path.join(processed_dir, "metadata")
    # Retrieve metadata fpaths in raw directory
    raw_metadata_fpaths = glob.glob(os.path.join(raw_metadata_dir, "*.yml"))
    # Copy all metadata yml files into the "processed" folder
    for raw_metadata_fpath in raw_metadata_fpaths:
        # Check if is a files
        if os.path.isfile(raw_metadata_fpath):
            metadata_fname = os.path.basename(raw_metadata_fpath)
            processed_metadata_fpath = os.path.join(
                processed_metadata_dir, metadata_fname
            )
            try:
                # Copy every file
                shutil.copy(raw_metadata_fpath, processed_metadata_fpath)
                msg = f"{metadata_fname} copied into {processed_metadata_dir}."
                logger.info(msg)
            except (Exception) as e:
                msg = f"Something went wrong when copying {metadata_fname} into {processed_metadata_dir}.\n The error is: {e}."
                logger.error(msg)
        else:
            msg = f"Cannot copy {metadata_fname} into {processed_metadata_dir}."
            logger.error(msg)
            raise ValueError(msg)
    metadata_fnames = [
        os.path.basename(metadata_fpath) for metadata_fpath in raw_metadata_fpaths
    ]
    msg = f"The metadata of stations ({metadata_fnames}) have been copied into {processed_metadata_dir}."
    logger.info(msg)


####--------------------------------------------------------------------------.
#### L0 processing directory structure


def create_directory_structure(raw_dir: str, processed_dir: str) -> None:
    """Create directory structure for L0A and L0B processing.

    Parameters
    ----------
    raw_dir : str
        Path of the raw directory
    processed_dir : str
        Path of the processed directory

    Raises
    ------
    FileNotFoundError
        Error is folder structure can not be created.
    """

    # -----------------------------------------------------.
    #### Create metadata folder inside processed_dir
    check_metadata_dir(processed_dir)
    copy_metadata_from_raw_dir(raw_dir, processed_dir)

    # -----------------------------------------------------.
    #### Create info folder inside processed_dir
    try:
        info_folder_path = os.path.join(processed_dir, "info")
        os.makedirs(info_folder_path)
        logger.debug(f"Created {info_folder_path}")
    except FileExistsError:
        logger.debug(f"Found {info_folder_path}")
        pass
    except (Exception) as e:
        msg = f"Can not create folder metadata inside <info_folder_path>. Error: {e}"
        logger.exception(msg)
        raise FileNotFoundError(msg)

    # -----------------------------------------------------.
    #### Create L0A folder inside processed_dir
    try:
        L0A_folder_path = os.path.join(processed_dir, "L0A")
        os.makedirs(L0A_folder_path)
        logger.debug(f"Created {L0A_folder_path}")
    except FileExistsError:
        logger.debug(f"Found {L0A_folder_path}")
        pass
    except (Exception) as e:
        msg = f"Can not create folder L0A inside {processed_dir}. Error: {e}"
        logger.exception(msg)
        raise FileNotFoundError(msg)

    # -----------------------------------------------------.
    #### Create L0B folder inside processed_dir
    try:
        L0B_folder_path = os.path.join(processed_dir, "L0B")
        os.makedirs(L0B_folder_path)
        logger.debug(f"Created {L0B_folder_path}")
    except FileExistsError:
        logger.debug(f"Found {L0B_folder_path}")
        pass
    except (Exception) as e:
        msg = f"Can not create folder L0B inside {processed_dir}. Error: {e}"
        logger.exception(msg)
        raise FileNotFoundError(msg)
    return


####--------------------------------------------------------------------------.
#### DISDRODB L0A Readers
def _read_L0A(
    fpath: str, lazy: bool = True, verbose: bool = False
) -> Union[pd.DataFrame, dd.DataFrame]:
    # Log
    msg = f" - Reading L0 Apache Parquet file at {fpath} started."
    log_info(logger, msg, verbose)
    # Read
    if lazy:
        df = dd.read_parquet(fpath)
    else:
        df = pd.read_parquet(fpath)
    # Log
    msg = f" - Reading L0 Apache Parquet file at {fpath} ended."
    log_info(logger, msg, verbose)
    return df


def read_L0A_dataframe(
    fpaths: Union[str, list],
    lazy: bool = True,
    verbose: bool = False,
    debugging_mode: bool = False,
) -> Union[pd.DataFrame, dd.DataFrame]:
    """Read DISDRODB L0A Apache Parquet file(s).

    Parameters
    ----------
    fpaths : str or list
        Either a list or a single filepath .
    verbose : bool
        Whether to print detailed processing information into terminal.
        The default is False.
    debugging_mode : bool
        If True, it reduces the amount of data to process.
        If fpaths is a list, it reads only the first 3 files
        The default is False.
    lazy : bool
        Whether to read the dataframe lazily with dask.
        If lazy=True, it returns a dask.dataframe.
        If lazy=False, it returns a pandas.DataFrame
        The default is True.

    Returns
    -------
    Union[pd.DataFrame,dd.DataFrame]
        Dataframe

    Raises
    ------
    TypeError
        Error if the reading fails
    """

    from disdrodb.L0.L0A_processing import concatenate_dataframe

    # ----------------------------------------
    # Check fpaths validity
    if not isinstance(fpaths, (list, str)):
        raise TypeError("Expecting fpaths to be a string or a list of strings.")
    # TODO:
    # - CHECK ENDS WITH .parquet
    # ----------------------------------------
    # If list of length 1, convert to string
    if isinstance(fpaths, list) and len(fpaths) == 1:
        fpaths = fpaths[0]
    # ---------------------------------------------------
    # If more than 1 fpath, read and concantenate first
    if isinstance(fpaths, list):
        if debugging_mode:
            fpaths = fpaths[0:3]  # select first 3 fpaths
        list_df = [_read_L0A(fpath, lazy=lazy, verbose=verbose) for fpath in fpaths]
        df = concatenate_dataframe(list_df, verbose=verbose, lazy=lazy)
    # Else read the single file
    else:
        df = _read_L0A(fpaths, lazy=lazy, verbose=verbose)
    # ---------------------------------------------------
    # Return dataframe
    return df


####---------------------------------------------------------------------------.


def check_L0_is_available(
    processed_dir: str, station_id: str, suffix: str = ""
) -> None:
    """Check if the Apache parquet file has been found

    Parameters
    ----------
    processed_dir : str
        Path of the processed directory
    station_id : int
        Id of the station.
    suffix : str, optional
        Suffix, by default ""

    Raises
    ------
    ValueError
        Check if the Apache parquet file has not been found
    """
    fpath = get_L0A_fpath(processed_dir, station_id, suffix=suffix)
    if not os.path.exists(fpath):
        msg = f"Need to run L0 processing first. The Apache Parquet file {fpath} is not available."
        logger.exception(msg)
        raise ValueError(msg)
    # Log
    msg = f"Found parquet file: {fpath}"
    logger.info(msg)


def read_L0_data(
    processed_dir: str,
    station_id: str,
    suffix: str = "",
    lazy: bool = True,
    verbose: bool = False,
    debugging_mode: bool = False,
) -> Union[pd.DataFrame, dd.DataFrame]:
    """Read L0 Apache Parquet into dataframe.


    Parameters
    ----------
    processed_dir : str
        Path of the processed directory
    station_id : int
        Id of the station.
    suffix : str, optional
        Suffix, by default ""
    lazy : bool, optional
        If True : Dask is used.
        If False : Pandas is used
        by default True
    verbose : bool, optional
        Wheter to verbose the processing.
        The default is False.
    debugging_mode : bool, optional
        If True, return just a subset of total rows, by default False

    Returns
    -------
    Union[pd.DataFrame,dd.DataFrame]
        Dataframe
    """

    # Check L0 is available
    check_L0_is_available(processed_dir, station_id, suffix=suffix)
    # Define fpath
    fpath = get_L0A_fpath(processed_dir, station_id, suffix=suffix)
    # Read L0A Apache Parquet file
    df = _read_L0A(fpath, lazy=lazy, verbose=verbose)
    # Subset dataframe if debugging_mode = True
    if debugging_mode:
        if not lazy:
            df = df.iloc[0:100, :]
        else:
            NotImplementedError
    return df


####--------------------------------------------------------------------------.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

# Retrieve sensor standards and configs

# -----------------------------------------------------------------------------.
import os
import yaml
import logging
import datetime

logger = logging.getLogger(__name__)

PRODUCT_VERSION = "V0"
SOFTWARE_VERSION = "V0"
EPOCH = "seconds since 1970-01-01 00:00:00"


# Notes:
# - L0A_encodings currently specify only the dtype. This could be expanded in the future.
# - disdrodb.configs ... the netcdf chunk size could be an option to be specified


def read_config_yml(sensor_name: str, filename: str) -> dict:
    """Read a config yaml file and return the dictionary.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.
    filename : str
        Name of the file.

    Returns
    -------
    dict
        Content of the config file.

    Raises
    ------
    ValueError
        Error if file does not exist.
    """

    # Get config path
    config_sensor_dir_path = get_configs_dir(sensor_name)
    fpath = os.path.join(config_sensor_dir_path, filename)
    # Check yaml file exists
    if not os.path.exists(fpath):
        msg = f"{filename} not available in {config_sensor_dir_path}"
        logger.exception(msg)
        raise ValueError(msg)
    # Open dictionary
    with open(fpath, "r") as f:
        d = yaml.safe_load(f)
    return d


def get_configs_dir(sensor_name: str) -> str:
    """Retrieve configs directory.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    str
        Config directory.

    Raises
    ------
    ValueError
        Error if the config directory does not exist.
    """

    dir_path = os.path.dirname(__file__)
    config_dir_path = os.path.join(dir_path, "configs")
    config_sensor_dir_path = os.path.join(config_dir_path, sensor_name)
    if not os.path.exists(config_sensor_dir_path):
        list_sensors = sorted(os.listdir(config_dir_path))
        print(f"Available sensor_name are {list_sensors}")
        raise ValueError(
            f"The config directory {config_sensor_dir_path} does not exist."
        )
    return config_sensor_dir_path


def get_available_sensor_name() -> sorted:
    """Get available names of sensors.

    Returns
    -------
    sorted
        Sorted list of the available sensors
    """

    dir_path = os.path.dirname(__file__)
    config_dir_path = os.path.join(dir_path, "configs")
    # TODO: here add checks that contains all required yaml file
    return sorted(os.listdir(config_dir_path))


def get_variables_dict(sensor_name: str) -> dict:
    """Get a dictionary containing the variable name of the sensor field numbers.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Variables names
    """

    return read_config_yml(sensor_name=sensor_name, filename="variables.yml")


def get_sensor_variables(sensor_name: str) -> list:
    """Get sensor variable names list.

    Parameters
    ----------
    sensor_name : str
         Name of the sensor.

    Returns
    -------
    list
        List of the variables values
    """

    return list(get_variables_dict(sensor_name).values())


def get_data_format_dict(sensor_name: str) -> dict:
    """Get a dictionary containing the data format of each sensor variable.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Data format of each sensor variable
    """

    return read_config_yml(sensor_name=sensor_name, filename="L0_data_format.yml")


def get_long_name_dict(sensor_name: str) -> dict:
    """Get a dictionary containing the long name of each sensor variable.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Long name of each sensor variable.
    """

    return read_config_yml(sensor_name=sensor_name, filename="variable_longname.yml")


def get_units_dict(sensor_name: str) -> dict:
    """Get a dictionary containing the unit of each sensor variable.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Unit of each sensor variable
    """

    return read_config_yml(sensor_name=sensor_name, filename="variable_units.yml")


def get_sensor_name_dict(sensor_name: str) -> dict:
    """Get a dictionary containing the description of each sensor variable.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Description of each sensor variable
    """

    d = read_config_yml(sensor_name=sensor_name, filename="variable_description.yml")
    return d


def get_diameter_bins_dict(sensor_name: str) -> dict:
    """Get dictionary with sensor_name diameter bins information.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        sensor_name diameter bins information
    """

    d = read_config_yml(sensor_name=sensor_name, filename="diameter_bins.yml")
    # TODO: Check dict contains center, bounds and width keys
    return d


def get_velocity_bins_dict(sensor_name: str) -> dict:
    """Get velocity with sensor_name diameter bins information.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Sensor_name diameter bins information
    """

    d = read_config_yml(sensor_name=sensor_name, filename="velocity_bins.yml")
    return d


def get_L0A_dtype(sensor_name: str) -> dict:
    """Get a dictionary containing the L0A dtype.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        L0A dtype
    """

    # Note: This function could extract the info from get_L0A_encodings_dict in future.
    d = read_config_yml(sensor_name=sensor_name, filename="L0A_encodings.yml")
    return d


def get_L0A_encodings_dict(sensor_name: str) -> dict:
    """Get a dictionary containing the L0A encodings

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        L0A encodings
    """

    # - L0A_encodings currently specify only the dtype. This could be expanded in the future.
    d = read_config_yml(sensor_name=sensor_name, filename="L0A_encodings.yml")
    return d


def get_L0B_encodings_dict(sensor_name: str) -> dict:
    """Get a dictionary containing the encoding to write L0B netCDFs.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Encoding to write L0B netCDFs
    """

    d = read_config_yml(sensor_name=sensor_name, filename="L0B_encodings.yml")

    # Ensure chunksize is a list
    for var in d.keys():
        if not isinstance(d[var]["chunksizes"], (list, type(None))):
            d[var]["chunksizes"] = [d[var]["chunksizes"]]

    # Sanitize encodings
    for var in d.keys():
        # Ensure contiguous=True if chunksizes is None
        if isinstance(d[var]["chunksizes"], type(None)) and not d[var]["contiguous"]:
            # These changes are required to enable netCDF writing
            d[var]["contiguous"] = True
            d[var]["fletcher32"] = False
            d[var]["zlib"] = False
            print(f"Set contiguous=True for variable {var} because chunksizes=None")
            print(f"Set fletcher32=False for variable {var} because contiguous=True")
            print(f"Set zlib=False for variable {var} because contiguous=True")
        # Ensure contiguous=False if chunksizes is not None
        if d[var]["contiguous"] and not isinstance(d[var]["chunksizes"], type(None)):
            d[var]["contiguous"] = False
            print(
                f"Set contiguous=False for variable {var} because chunksizes is defined!"
            )

    return d


def get_time_encoding() -> dict:
    """Create time encoding

    Returns
    -------
    dict
        Time encoding
    """
    encoding = {}
    encoding["units"] = EPOCH
    encoding["calendar"] = "proleptic_gregorian"
    return encoding


def set_DISDRODB_L0_attrs(ds, attrs: dict):
    """Define DISDRODB L0 attributes

    Parameters
    ----------
    ds : xarray dataset
        Dataset
    attrs : dict
        Attributes

    Returns
    -------
    xarray dataset
        Dataset
    """
    sensor_name = attrs["sensor_name"]
    # ----------------------------------
    # Set global attributes (from metadata)
    ds.attrs = attrs

    # ----------------------------------
    # Set variable attributes (from config standards)
    description_dict = get_sensor_name_dict(sensor_name)
    units_dict = get_units_dict(sensor_name)
    long_name_dict = get_long_name_dict(sensor_name)
    # data_format_dict = get_data_format_dict(sensor_name : str)
    for var in list(ds.data_vars):
        attrs_var = {}
        attrs_var["long_name"] = long_name_dict[var]
        attrs_var["description"] = description_dict[var]
        attrs_var["unit"] = units_dict[var]
        # TODO:
        # attrs['valid_min'] =  data_format_dict
        # attrs['valid_max'] =  data_format_dict
        ds[var].attrs = attrs_var

    # ----------------------------------
    # Set coordinate attributes
    # TODO
    # lat/lon - degrees N/E...
    # altitude - meters above sea level

    # ----------------------------------
    # Add DISDRODB processing info
    now = datetime.datetime.now()
    current_time = now.strftime("%Y-%m-%d %H:%M:%S")
    ds.attrs["disdrodb_processing_date"] = current_time
    ds.attrs["disdrodb_product_version"] = PRODUCT_VERSION
    ds.attrs["disdrodb_software_version"] = SOFTWARE_VERSION
    ds.attrs["disdrodb_product_level"] = "L0"
    # ds.attrs['disdrodb_id'] = # TODO XXXX

    return ds


####-------------------------------------------------------------------------.
#############################################
#### Get diameter and velocity bins info ####
#############################################


def get_diameter_bin_center(sensor_name: str) -> list:
    """Get diameter bin center.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor

    Returns
    -------
    list
        Diameter bin center
    """

    diameter_dict = get_diameter_bins_dict(sensor_name)
    diameter_bin_center = list(diameter_dict["center"].values())
    return diameter_bin_center


def get_diameter_bin_lower(sensor_name: str) -> list:
    """Get diameter bin lower bound.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor

    Returns
    -------
    list
        Diameter bin lower bound
    """

    diameter_dict = get_diameter_bins_dict(sensor_name)
    lower_bounds = [v[0] for v in diameter_dict["bounds"].values()]
    return lower_bounds


def get_diameter_bin_upper(sensor_name: str) -> list:
    """Get diameter bin upper bound.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor

    Returns
    -------
    list
        Diameter bin upper bound
    """

    diameter_dict = get_diameter_bins_dict(sensor_name)
    upper_bounds = [v[1] for v in diameter_dict["bounds"].values()]
    return upper_bounds


def get_diameter_bin_width(sensor_name: str) -> list:
    """Get diameter bin width.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor

    Returns
    -------
    list
        Diameter bin width
    """

    diameter_dict = get_diameter_bins_dict(sensor_name)
    diameter_bin_width = list(diameter_dict["width"].values())
    return diameter_bin_width


def get_velocity_bin_center(sensor_name: str) -> list:
    """Get velocity bin center.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor

    Returns
    -------
    list
        Velocity bin center
    """

    velocity_dict = get_velocity_bins_dict(sensor_name)
    if velocity_dict is not None:
        velocity_bin_center = list(velocity_dict["center"].values())
    else:
        return None
    return velocity_bin_center


def get_velocity_bin_lower(sensor_name: str) -> list:
    """Get velocity bin lower bound.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor

    Returns
    -------
    list
        Velocity bin lower bound.
    """

    velocity_dict = get_velocity_bins_dict(sensor_name)
    if velocity_dict is not None:
        lower_bounds = [v[0] for v in velocity_dict["bounds"].values()]
    else:
        return None
    return lower_bounds


def get_velocity_bin_upper(sensor_name: str) -> list:
    """Get velocity bin upper bound.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor

    Returns
    -------
    list
        Velocity bin upper bound
    """

    velocity_dict = get_velocity_bins_dict(sensor_name)
    if velocity_dict is not None:
        upper_bounds = [v[1] for v in velocity_dict["bounds"].values()]
    else:
        return None
    return upper_bounds


def get_velocity_bin_width(sensor_name: str) -> list:
    """Get velocity bin width.

    Parameters
    ----------
    sensor_name : str
        Name of the sensor

    Returns
    -------
    list
        Velocity bin width
    """

    velocity_dict = get_velocity_bins_dict(sensor_name)
    if velocity_dict is not None:
        velocity_bin_width = list(velocity_dict["width"].values())
    else:
        return None
    return velocity_bin_width


def get_raw_field_nbins(sensor_name: str) -> dict:
    """Get the raw field number of bins

    Parameters
    ----------
    sensor_name : str
        Name of the sensor.

    Returns
    -------
    dict
        Field definition.
    """

    diameter_dict = get_diameter_bins_dict(sensor_name)
    velocity_dict = get_velocity_bins_dict(sensor_name)
    n_d = len(diameter_dict["center"])
    # For instruments measuring size and velocity (i.e. OTT Parsivel, ThiesLPM)
    if velocity_dict is not None:
        n_v = len(velocity_dict["center"])
        nbins_dict = {
            "raw_drop_concentration": n_d,
            "raw_drop_average_velocity": n_v,
            "raw_drop_number": n_d * n_v,
        }
    # For instruments measuring only size (i.e. RD80)
    else:
        nbins_dict = {
            "raw_drop_number": n_d,
        }
    return nbins_dict


def get_raw_field_dim_order(sensor_name: str) -> dict:
    """Get dimention order  dictionnary

    Parameters
    ----------
    sensor_name : str
        Name of the sensor

    Returns
    -------
    dict
        Dimention order  dictionnary

    Raises
    ------
    NotImplementedError
        Name of the sensor not implemented.
    """
    # TODO: this should go into a config file ...
    # TODO: also think to set dimensions as diameter and velocity ... TO DISCUSS
    if sensor_name in ["OTT_Parsivel", "OTT_Parsivel2", "ThiesLPM"]:
        dim_dict = {
            "raw_drop_concentration": ["diameter_bin_center"],
            "raw_drop_average_velocity": ["velocity_bin_center"],
            "raw_drop_number": ["diameter_bin_center", "velocity_bin_center"],
        }
    elif sensor_name in ["RD80"]:
        dim_dict = {"raw_drop_number": ["diameter_bin_center"]}
    else:
        raise NotImplementedError()
    return dim_dict


def get_raw_spectrum_ndims(sensor_name: str):
    encoding_dict = get_L0B_encodings_dict(sensor_name)
    ndim = len(encoding_dict["raw_drop_number"]["chunksizes"]) - 1
    return ndim


# -----------------------------------------------------------------------------.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jan  1 17:18:04 2022

@author: ghiggi
"""
import os
import subprocess
from disdrodb.pipeline.utils_cmd import get_reader_cmd

# You need to set the disdrodb repo path in your .bashrc
# export PYTHONPATH="${PYTHONPATH}:/home/ghiggi/Projects/disdrodb"
# You need to activate the disdrodb envirnment: conda activate disdrodb

# -----------------------------------------------------------------------------.
#### Define campaign dictionary
EPFL_dict = {
    "LOCARNO_2018": "reader_LOCARNO_2018.py",
    "PARSIVEL_2007": "reader_PARSIVEL_2007.py",
    "GENEPI_2007": "reader_GENEPI_2007.py",  # Asked to discard into metadata
    "EPFL_ROOF_2008_1": "reader_EPFL_ROOF_2008_1.py",
    "EPFL_ROOF_2008_2": "reader_EPFL_ROOF_2008_2.py",
    "EPFL_ROOF_2011": "reader_EPFL_ROOF_2011.py",
    "EPFL_ROOF_2012": "reader_EPFL_ROOF_2012.py",
    "EPFL_2009": "reader_EPFL_2009.py",
    "DAVOS_2009_2011": "reader_DAVOS_2009_2011.py",
    "HPICONET_2010": "reader_HPICONET_2010.py",
    "COMMON_2011": "reader_COMMON_2011.py",
    "RIETHOLZBACK_2011": "reader_RIETHOLZBACK_2011.py",
    "HYMEX": "reader_HYMEX.py",
    "HYMEX": "reader_HYMEX.py",
    "PARADISO_2014": "reader_PARADISO_2014.py",
    "SAMOYLOV_2017_2019": "reader_SAMOYLOV_2017_2019.py",
    "LOCARNO_2018": "reader_LOCARNO_2018.py",
    "PLATO_2019": "reader_PLATO_2019.py",
    "DAVOS_2009": "reader_DAVOS_2009.py",
    "EPFL_ROOF_2010": "reader_EPFL_ROOF_2010.py",
    "PARADISO_2014": "reader_PARADISO_2014.py",
    "RACLETS_2019": "reader_RACLETS_2019.py",
    "SAMOYLOV_2017_2019": "reader_SAMOYLOV_2017_2019.py",
}

#### Define filepaths
reader_dir = "/ltenas3/0_Projects/disdrodb/disdrodb/readers/EPFL"  # TODO: this should change to the current package
reader_dir = "~/disdrodb/disdrodb/L0/readers/EPFL"

raw_base_dir = "/ltenas3/0_Data/DISDRODB/Raw/EPFL"

processed_base_dir = "/ltenas3/0_Data/DISDRODB/Processed/EPFL"
processed_base_dir = "/tmp/DISDRODB/Processed/EPFL"
processed_base_dir = "/home/kcandolf/tmp/DISDRODB/Processed/EPFL"


#### Processing settings
l0a_processing = True
l0b_processing = True
keep_l0a = True
force = True
verbose = True
debugging_mode = True
lazy = True
single_netcdf = True


#### Process all campaigns
for campaign_name in EPFL_dict.keys():
    reader_filepath = os.path.join(reader_dir, EPFL_dict[campaign_name])
    cmd = get_reader_cmd(
        reader_filepath=reader_filepath,
        raw_dir=os.path.join(raw_base_dir, campaign_name),
        processed_dir=os.path.join(processed_base_dir, campaign_name),
        l0a_processing=l0a_processing,
        l0b_processing=l0b_processing,
        keep_l0a=keep_l0a,
        single_netcdf=single_netcdf,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=lazy,
    )

    subprocess.run(cmd, shell=True)
    # os.system(cmd)

# -----------------------------------------------------------------------------.
# TODO:
# --> Useful to test changes to code do not crash other reader
# --> debuggin_mode=True to speed up tests ;)

# -----------------------------------------------------------------------------.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 15 09:48:21 2022

@author: kimbo
"""
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

from disdrodb.L0 import run_L0


from disdrodb.L0.L0_processing import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    l0a_processing=True,
    l0b_processing=True,
    keep_l0a=False,
    force=False,
    verbose=False,
    debugging_mode=False,
    lazy=True,
    single_netcdf=True,
):

    ####----------------------------------------------------------------------.
    ###########################
    #### CUSTOMIZABLE CODE ####
    ###########################
    #### - Define raw data headers
    # Notes
    # - In all files, the datalogger voltage hasn't the delimeter,
    #   so need to be split to obtain datalogger_voltage and rainfall_rate_32bit

    # These are the variables included in the datasets:

    # var			full name										units

    # Time		time of the record								Y-m-d hh:mm:ss
    # Event		event ID 										(factor)
    # ID			disdromter ID 									(factor: T1, T2, P1, P2)
    # Serial		disdrometer serial number						(factor)
    # Type		disdrometer type 								(factor: Thi, Par)
    # Mast		mast ID											(factor: 1, 2)
    # NP_meas		number of particles detected					(-)
    # R_meas		rainfall intensity, as outputted by the device	mm h−1
    # Z_meas		radar reflectivity, as outputted by the device	dB mm6 m−3
    # E_meas		erosivity, as outputted by the device			J m−2 mm−1
    # Pcum_meas	cumulative rainfall amount						mm
    # Ecum_meas	cumulative kinetic energy						J m−2 mm−1
    # NP			number of particles detected					(-)
    # ND			particle density								m−3 mm−1
    # R			rainfall intensity								mm h−1
    # P			rainfall amount									mm
    # Z			radar reflectivity								dB mm6 m−3
    # M			water content									gm−3
    # E			kinetic energy									J m−2 mm−1
    # Pcum		cumulative rainfall amount						mm
    # Ecum		cumulative kinetic energy						J m−2 mm−1
    # D10			drop diameter’s 10th percentile					mm
    # D25			drop diameter’s 25th percentile					mm
    # D50			drop diameter’s 50th percentile					mm
    # D75			drop diameter’s 75th percentile					mm
    # D90			drop diameter’s 90th percentile					mm
    # Dm			mean drop diameter								mm
    # V10			drop velocity’s 10th percentile					m s−1
    # V25			drop velocity’s 25th percentile					m s−1
    # V50			drop velocity’s 50th percentile					m s−1
    # V75			drop velocity’s 75th percentile					m s−1
    # V90			drop velocity’s 90th percentile					m s−1
    # Vm			mean drop velocity								m s−1

    column_names = [
        "time",
        "id",
        "disdromter_ID",  # to_drop
        "disdrometer_serial",  # to_drop
        "disdrometer_type",  # to_drop
        "mast_ID",  # to_drop
        "number_particles_meas",  # to_drop
        "rainfall_rate_32bit_meas",  # to_drop
        "reflectivity_32bit_meas",  # to_drop
        "mor_visibility_meas",  # to_drop
        "rainfall_accumulated_32bit_meas",  # to_drop
        "rain_kinetic_energy_meas",
        "number_particles",
        "rainfall_rate_32bit",
        "reflectivity_32bit",
        "temp",  # I think is mor_visibility, but not sure about this, give error because values are like: 0.00386755693894061
        "rainfall_accumulated_32bit",
        "rain_kinetic_energy",
        "D10",  # I don't know what to do with this
        "D25",  # I don't know what to do with this
        "D50",  # I don't know what to do with this
        "D75",  # I don't know what to do with this
        "D90",  # I don't know what to do with this
        "Dm",  # I don't know what to do with this
        "V10",  # I don't know what to do with this
        "V25",  # I don't know what to do with this
        "V50",  # I don't know what to do with this
        "V75",  # I don't know what to do with this
        "V90",  # I don't know what to do with this
        "Vm",  # I don't know what to do with this
    ]

    ##------------------------------------------------------------------------.
    #### - Define reader options
    reader_kwargs = {}

    # - Define delimiter
    reader_kwargs["delimiter"] = ","

    # - Avoid first column to become df index !!!
    reader_kwargs["index_col"] = False

    # - Define behaviour when encountering bad lines
    reader_kwargs["on_bad_lines"] = "skip"

    # - Define reader engine
    #   - C engine is faster
    #   - Python engine is more feature-complete
    reader_kwargs["engine"] = "python"

    # - Define on-the-fly decompression of on-disk data
    #   - Available: gzip, bz2, zip
    reader_kwargs["compression"] = "infer"

    # - Strings to recognize as NA/NaN and replace with standard NA flags
    #   - Already included: ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’,
    #                       ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’,
    #                       ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
    reader_kwargs["na_values"] = ["na", "", "error", "NA", "-.-"]

    # - Define max size of dask dataframe chunks (if lazy=True)
    #   - If None: use a single block for each file
    #   - Otherwise: "<max_file_size>MB" by which to cut up larger files
    reader_kwargs["blocksize"] = None  # "50MB"

    # Cast all to string
    reader_kwargs["dtype"] = str

    # Skip first row as columns names
    reader_kwargs["header"] = None

    ##------------------------------------------------------------------------.
    #### - Define facultative dataframe sanitizer function for L0 processing
    # - Enable to deal with bad raw data files
    # - Enable to standardize raw data files to L0 standards  (i.e. time to datetime)
    df_sanitizer_fun = None

    def df_sanitizer_fun(df, lazy=False):
        # Import dask or pandas
        if lazy:
            import dask.dataframe as dd
        else:
            import pandas as dd

        # # Drop useless columns
        df = df.drop(
            columns=[
                "id",
                "disdromter_ID",  # to_drop
                "disdrometer_serial",  # to_drop
                "disdrometer_type",  # to_drop
                "mast_ID",  # to_drop
                "number_particles_meas",  # to_drop
                "rainfall_rate_32bit_meas",  # to_drop
                "reflectivity_32bit_meas",  # to_drop
                "temp",  # I think is mor_visibility, but not sure about this, give error because values are like: 0.00386755693894061
                "mor_visibility_meas",  # to_drop
                "rainfall_accumulated_32bit_meas",  # to_drop
                "rain_kinetic_energy_meas",
                "D10",  # I don't know what to do with this
                "D25",  # I don't know what to do with this
                "D50",  # I don't know what to do with this
                "D75",  # I don't know what to do with this
                "D90",  # I don't know what to do with this
                "Dm",  # I don't know what to do with this
                "V10",  # I don't know what to do with this
                "V25",  # I don't know what to do with this
                "V50",  # I don't know what to do with this
                "V75",  # I don't know what to do with this
                "V90",  # I don't know what to do with this
                "Vm",  # I don't know what to do with this
            ]
        )

        # - Convert time column to datetime
        df["time"] = dd.to_datetime(df["time"], format="%Y-%m-%d %H:%M:%S")

        return df

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files in raw_dir/data/<station_id>
    files_glob_pattern = "*.csv*"

    ####----------------------------------------------------------------------.
    #### - Create L0 products
    run_L0(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        l0a_processing=l0a_processing,
        l0b_processing=l0b_processing,
        keep_l0a=keep_l0a,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=lazy,
        single_netcdf=single_netcdf,
        # Custom arguments of the reader
        files_glob_pattern=files_glob_pattern,
        column_names=column_names,
        reader_kwargs=reader_kwargs,
        df_sanitizer_fun=df_sanitizer_fun,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 24 20:30:51 2022

@author: ghiggi
"""
import os
import time
import glob
import shutil
from urllib import response
import click
import logging

# Directory
from disdrodb.L0.io import (
    check_directories,
    get_campaign_name,
    create_directory_structure,
    get_L0A_dir,
    get_L0A_fpath,
    get_L0B_fpath,
    read_L0A_dataframe,
)

# Metadata
from disdrodb.L0.metadata import read_metadata

# Standards
from disdrodb.L0.check_standards import check_sensor_name, check_L0A_standards

# L0A_processing
from disdrodb.L0.L0A_processing import (
    get_file_list,
    read_L0A_raw_file_list,
    write_df_to_parquet,
)

# L0B_processing
from disdrodb.L0.L0B_processing import (
    create_L0B_from_L0A,
    write_L0B,
    create_summary_statistics,
)

# Logger
from disdrodb.utils.logger import create_L0_logger, close_logger
from disdrodb.utils.logger import log_info, log_warning

logger = logging.getLogger(__name__)

# ------------------------------------------------------------------.
# Consistency choice
# TODO:
# - Add verbose and logs to disdrodb.io function !!!

# -----------------------------------------------------------------------------.
def click_L0_readers_options(function: object):
    """Define click command line parameters.

    Parameters
    ----------
    function : object
        Function.
    """
    function = click.option(
        "-s",
        "--single_netcdf",
        type=bool,
        show_default=True,
        default=True,
        help="Produce single netCDF",
    )(function)
    function = click.option(
        "-l",
        "--lazy",
        type=bool,
        show_default=True,
        default=True,
        help="Use dask if lazy=True",
    )(function)
    function = click.option(
        "-d",
        "--debugging_mode",
        type=bool,
        show_default=True,
        default=False,
        help="Switch to debugging mode",
    )(function)
    function = click.option(
        "-v", "--verbose", type=bool, show_default=True, default=False, help="Verbose"
    )(function)
    function = click.option(
        "-f",
        "--force",
        type=bool,
        show_default=True,
        default=False,
        help="Force overwriting",
    )(function)
    function = click.option(
        "-k",
        "--keep_l0a",
        type=bool,
        show_default=True,
        default=True,
        help="Whether to keep the L0A Parquet file",
    )(function)
    function = click.option(
        "-l0b",
        "--l0b_processing",
        type=bool,
        show_default=True,
        default=True,
        help="Perform L0B processing",
    )(function)
    function = click.option(
        "-l0a",
        "--l0a_processing",
        type=bool,
        show_default=True,
        default=True,
        help="Perform L0A processing",
    )(function)
    function = click.argument("processed_dir", metavar="<processed_dir>")(function)
    function = click.argument(
        "raw_dir", type=click.Path(exists=True), metavar="<raw_dir>"
    )(function)

    return function


# -----------------------------------------------------------------------------.
def run_L0(
    # Arguments custom to each reader
    column_names,
    reader_kwargs,
    files_glob_pattern,
    df_sanitizer_fun,
    raw_dir,
    processed_dir,
    # Arguments designing the processing type
    l0a_processing=True,
    l0b_processing=True,
    keep_l0a=True,
    force=False,
    verbose=False,
    debugging_mode=False,
    lazy=True,
    single_netcdf=True,
):
    """Core function to process raw data files to L0A and L0B format.

    Parameters
    ----------

    column_names : list
        Column names to be assigned to the dataframe created after reading a raw file.
        This argument is allowed to contains column names not agreeing with the
        DISDRODB standard. However, it's important that non-standard columns are
        dropped during the application of the `df_sanitizer_fun` function.
    reader_kwargs : dict
        Dictionary of arguments to be passed to `pd.read_csv` or `dd.read_csv` to
        tailor the reading of the raw file.
    files_glob_pattern: str
        It indicates the glob pattern to search for the raw data files within
        the directory path <raw_dir>/data/<station_id>.
    df_sanitizer_fun : function
        It's a function detailing the custom preprocessing of each single raw data files
        in order to meet the DISDRODB standards.
        The function must have the following definition and return a dataframe:
            def df_sanitizer_fun(df, lazy=False)
                # Import dask or pandas depending on lazy argument
                if lazy:
                    import dask.dataframe as dd
                else:
                    import pandas as dd
                # Custom processing
                pass
                # Return the dataframe agreeing with the DISDRODB standards
                return df
    raw_dir : str
        The directory path where all the raw content of a specific campaign is stored.
        The path must have the following structure:
            <...>/DISDRODB/Raw/<data_source>/<campaign_name>'.
        Inside the raw_dir directory, it is required to adopt the following structure:
        - /data/<station_id>/<raw_files>
        - /metadata/<station_id>.yaml
        Important points:
        - For each <station_id> there must be a corresponding YAML file in the metadata subfolder.
        - The <campaign_name> must semantically match between:
           - the raw_dir and processed_dir directory paths;
           - with the key 'campaign_name' within the metadata YAML files.
        - The campaign_name are expected to be UPPER CASE.
    processed_dir : str
        The desired directory path for the processed DISDRODB L0A and L0B products.
        The path should have the following structure:
            <...>/DISDRODB/Processed/<data_source>/<campaign_name>'
        For testing purpose, this function exceptionally accept also a directory path simply ending
        with <campaign_name> (i.e. /tmp/<campaign_name>).
    l0a_processing : bool
      Whether to launch processing to generate L0A Apache Parquet file(s) from raw data.
      The default is True.
    l0b_processing : bool
      Whether to launch processing to generate L0B netCDF4 file(s) from L0A data.
      The default is True.
    keep_l0a : bool
        Whether to keep the L0A files after having generated the L0B netCDF products.
        The default is False.
    force : bool
        If True, overwrite existing data into destination directories.
        If False, raise an error if there are already data into destination directories.
        The default is False.
    verbose : bool
        Whether to print detailed processing information into terminal.
        The default is False.
    debugging_mode : bool
        If True, it reduces the amount of data to process.
        - For L0A processing, it processes just 3 raw data files.
        - For L0B processing, it takes a small subset of the L0A Apache Parquet dataframe.
        The default is False.
    lazy : bool
        Whether to perform processing lazily with dask.
        If lazy=True, it employed dask.array and dask.dataframe.
        If lazy=False, it employed pandas.DataFrame and numpy.array.
        The default is True.
    single_netcdf : bool
        Whether to concatenate all raw files into a single L0B netCDF file.
        If single_netcdf=True, all raw files will be saved into a single L0B netCDF file.
        If single_netcdf=False, each raw file will be converted into the corresponding L0B netCDF file.
        The default is True.

    """

    t_i_script = time.time()
    # -------------------------------------------------------------------------.
    # Initial directory checks
    raw_dir, processed_dir = check_directories(raw_dir, processed_dir, force=force)

    # Retrieve campaign name
    campaign_name = get_campaign_name(raw_dir)

    # -------------------------------------------------------------------------.
    # Define logging settings
    logger = create_L0_logger(processed_dir, campaign_name)

    # -------------------------------------------------------------------------.
    # Create directory structure
    create_directory_structure(raw_dir, processed_dir)

    # -------------------------------------------------------------------------.
    #### Loop over station_id directory and process the files
    list_stations_id = os.listdir(os.path.join(raw_dir, "data"))

    # station_id = list_stations_id[0]
    for station_id in list_stations_id:
        # ---------------------------------------------------------------------.
        logger.info(f" - Processing of station_id {station_id} has started")
        # ---------------------------------------------------------------------.
        # Retrieve metadata
        attrs = read_metadata(raw_dir=raw_dir, station_id=station_id)
        # Retrieve sensor name
        sensor_name = attrs["sensor_name"]
        check_sensor_name(sensor_name)

        # ---------------------------------------------------------------------.
        ########################
        #### L0A processing ####
        ########################
        if l0a_processing:
            # Start L0 processing
            t_i_station = time.time()
            msg = " - L0A processing of station_id {} has started.".format(station_id)
            if verbose:
                print(msg)
            logger.info(msg)

            # -----------------------------------------------------------------.
            #### - List files to process
            glob_pattern = os.path.join("data", station_id, files_glob_pattern)
            file_list = get_file_list(
                raw_dir=raw_dir,
                glob_pattern=glob_pattern,
                verbose=verbose,
                debugging_mode=debugging_mode,
            )

            # -----------------------------------------------------------------.
            #### - If single_netcdf = True, ensure loop over all files only once
            if single_netcdf:
                file_list = [file_list]

            # -----------------------------------------------------------------.
            #### - Loop over all files
            # - It loops only once if single_netcdf=True
            for filepath in file_list:
                ##------------------------------------------------------.
                # Define file suffix
                if single_netcdf:
                    file_suffix = ""
                else:
                    # Get file name without file extensions
                    t_i_file = time.time()
                    file_suffix = os.path.basename(filepath).split(".")[0]
                    logger.info(
                        f"L0A processing of raw file {file_suffix} has started."
                    )

                ##------------------------------------------------------.
                #### - Read all raw data files into a dataframe
                df = read_L0A_raw_file_list(
                    file_list=filepath,
                    column_names=column_names,
                    reader_kwargs=reader_kwargs,
                    df_sanitizer_fun=df_sanitizer_fun,
                    lazy=lazy,
                    sensor_name=sensor_name,
                    verbose=verbose,
                )

                ##------------------------------------------------------.
                #### - Write to Parquet
                fpath = get_L0A_fpath(processed_dir, station_id, suffix=file_suffix)
                write_df_to_parquet(df=df, fpath=fpath, force=force, verbose=verbose)

                ##------------------------------------------------------.
                #### - Check L0 file respects the DISDRODB standards
                check_L0A_standards(
                    fpath=fpath, sensor_name=sensor_name, verbose=verbose
                )

                ##------------------------------------------------------.
                # Delete temp variables
                del df

                ##------------------------------------------------------.
                if not single_netcdf:
                    # End L0 processing for a single raw file
                    t_f = time.time() - t_i_file
                    msg = " - L0A processing of {} ended in {:.2f}s".format(
                        file_suffix, t_f
                    )
                    log_info(logger, msg, verbose)

            ##------------------------------------------------------.
            # End L0 processing for the station
            t_f = time.time() - t_i_station
            msg = " - L0A processing of station_id {} ended in {:.2f}s".format(
                station_id, t_f
            )
            log_info(logger, msg, verbose)

        # ------------------------------------------------------------------.
        ########################
        #### L0B processing ####
        ########################
        if l0b_processing:
            # Start L1 processing
            t_i = time.time()
            msg = " - L0B processing of station_id {} has started.".format(station_id)
            if verbose:
                print(msg)
            logger.info(msg)
            ##----------------------------------------------------------------.
            # Get station L0A directory
            L0A_dir_path = get_L0A_dir(processed_dir, station_id)
            file_list = glob.glob(os.path.join(L0A_dir_path, "*.parquet"))
            n_files = len(file_list)
            if n_files == 0:
                msg = f"No L0A Apache Parquet file is available in {L0A_dir_path}. Run L0A processing first."
                logger.error(msg)
                raise ValueError(msg)

            ##----------------------------------------------------------------.
            # Checks for single_netcdf=True
            if single_netcdf:
                # Enclose into a list to loop over only once
                file_list = [file_list]
                if n_files != 1:
                    msg = "If single_netcdf=True, DISDRODB would typically expect only a single L0A Apache Parquet file in {L0A_dir_path}."
                    log_warning(logger, msg, verbose)

            ##----------------------------------------------------------------.
            # Loop over all files
            for filepath in file_list:
                ##------------------------------------------------------.
                # Define file suffix
                if single_netcdf:
                    file_suffix = ""
                else:
                    # Get file name without file extensions
                    t_i_file = time.time()
                    file_suffix = os.path.basename(filepath).split(".")[0]
                    logger.info(
                        f"L0A processing of raw file {file_suffix} has started."
                    )
                ##------------------------------------------------------.
                # Read L0A dataframes
                df = read_L0A_dataframe(
                    filepath, lazy=lazy, verbose=verbose, debugging_mode=debugging_mode
                )

                # -----------------------------------------------------------------.
                #### - Create xarray Dataset
                ds = create_L0B_from_L0A(df=df, attrs=attrs, lazy=lazy, verbose=verbose)

                # -----------------------------------------------------------------.
                #### - Write L0B netCDF4 dataset
                fpath = get_L0B_fpath(ds, processed_dir, station_id)
                write_L0B(ds, fpath=fpath, sensor_name=sensor_name)

                # -----------------------------------------------------------------.
                if not single_netcdf:
                    # End L0B processing for a single L0A file
                    t_f = time.time() - t_i_file
                    msg = " - L0B processing of {} ended in {:.2f}s".format(
                        file_suffix, t_f
                    )
                    log_info(logger, msg, verbose)

            # -----------------------------------------------------------------.
            #### - Compute L0B summary statics (if single_netcdf=True)
            if single_netcdf:
                create_summary_statistics(
                    ds,
                    processed_dir=processed_dir,
                    station_id=station_id,
                    sensor_name=sensor_name,
                )

            # -----------------------------------------------------------------.
            # End L0B processing
            t_f = time.time() - t_i
            msg = " - L0B processing of station_id {} ended in {:.2f}s".format(
                station_id, t_f
            )
            log_info(logger, msg, verbose)

            # -----------------------------------------------------------------.
        # ---------------------------------------------------------------------.
    # Remove L0A directory if keep_L0A = False
    if not keep_l0a:
        shutil.rmtree(os.path.join(processed_dir, "L0A"))

    # -------------------------------------------------------------------------.
    # End of L0B processing for all stations
    t_f = time.time() - t_i_script
    msg = " - L0 processing of stations {} ended in {:.2f} minutes".format(
        list_stations_id, t_f / 60
    )

    # -------------------------------------------------------------------------.
    # Final logs
    logger.info("---")
    msg = "### Script finish ###"
    log_info(logger, msg, verbose)
    close_logger(logger)


def get_available_readers() -> dict:
    """Returns the readers description included into the current release of DISDRODB.

    Returns
    -------
    dict
        The dictionary has the following schema {"data_source":{"campaign_name":"reader file path"}}
    """
    # current file path
    lo_folder_path = os.path.dirname(__file__)

    # readers folder path
    reader_folder_path = os.path.join(lo_folder_path, "readers")

    # list of readers folder
    list_of_reader_folder = [
        f.path for f in os.scandir(reader_folder_path) if f.is_dir()
    ]

    # create dictionary
    dict_reader = {}
    for path_folder in list_of_reader_folder:
        data_source = os.path.basename(path_folder)
        dict_reader[data_source] = {}
        for path_python_file in [
            f.path
            for f in os.scandir(path_folder)
            if f.is_file() and f.path.endswith(".py")
        ]:
            reader_name = (
                os.path.basename(path_python_file)
                .replace("reader_", "")
                .replace(".py", "")
            )
            dict_reader[data_source][reader_name] = path_python_file

    return dict_reader


def check_data_source(data_source: str) -> str:
    """Check if the provided data source exists within the available readers.
    Please run get_available_readers() to get the list of all available reader

    Parameters
    ----------
    data_source : str
        Data source name  - Institution name (when campaign data spans more than 1 country) or country (when all campaigns (or sensor networks) are inside a given country)

    Returns
    -------
    str
        if data source exists : retrurn the correct data source name
        if data source does not exist : error

    Raises
    ------
    ValueError
        Error if the data source name provided has not been found.
    """

    dict_all_readers = get_available_readers()

    correct_data_source_list = list(
        set(dict_all_readers.keys()).intersection([data_source, data_source.upper()])
    )

    if correct_data_source_list:
        correct_data_source = correct_data_source_list[0]
    else:
        msg = (
            f"Data source {data_source} has not been found within the available readers"
        )
        logger.exception(msg)
        raise ValueError(msg)

    return correct_data_source


def get_available_readers_by_data_source(data_source: str) -> dict:
    """Return the available readers by data source

    Parameters
    ----------
    data_source : str
        Data source name - Institution name (when campaign data spans more than 1 country) or country (when all campaigns (or sensor networks) are inside a given country)

    Returns
    -------
    dict
        Dictionary that conatins the campaigns for the requested data source.

    """

    correct_data_source = check_data_source(data_source)

    if correct_data_source:
        dict_data_source = get_available_readers().get(correct_data_source)

    return dict_data_source


def check_reader_name(data_source: str, reader_name: str) -> str:
    """Check if the provided data source exists and reader names exists within the available readers.
    Please run get_available_readers() to get the list of all available reader

    Parameters
    ----------
    data_source : str
        Data source name - Institution name (when campaign data spans more than 1 country) or country (when all campaigns (or sensor networks) are inside a given country)
    reader_name : str
        Campaign name

    Returns
    -------
    str
        If True : returns the reader name
        If False : Error - return None

    Raises
    ------
    ValueError
        Error if the reader name provided for the campaign has not been found.
    """

    correct_data_source = check_data_source(data_source)

    if correct_data_source:
        dict_reader_names = get_available_readers_by_data_source(correct_data_source)

        correct_reader_name_list = list(
            set(dict_reader_names.keys()).intersection(
                [reader_name, reader_name.upper()]
            )
        )

        if correct_reader_name_list:
            correct_reader_name = correct_reader_name_list[0]
        else:
            msg = (
                f"Reader {reader_name} has not been found within the available readers"
            )
            logger.exception(msg)
            raise ValueError(msg)

    return correct_reader_name


def get_reader(data_source: str, reader_name: str) -> object:
    """Returns the reader function based on input parameters

    Parameters
    ----------
    data_source : str
        Institution name (when campaign data spans more than 1 country) or country (when all campaigns (or sensor networks) are inside a given country)
    reader_name : str
        Campaign name

    Returns
    -------
    object
        The reader() function

    """

    corrcet_data_source = check_data_source(data_source)
    correct_reader_name = check_reader_name(data_source, reader_name)

    if correct_reader_name:
        full_name = (
            f"disdrodb.L0.readers.{corrcet_data_source}.{correct_reader_name}.reader"
        )
        module_name, unit_name = full_name.rsplit(".", 1)
        my_reader = getattr(__import__(module_name, fromlist=[""]), unit_name)

    return my_reader


def run_reader(
    data_source: str,
    reader_name: str,
    raw_dir: str,
    processed_dir: str,
    l0a_processing: bool = True,
    l0b_processing: bool = True,
    keep_l0a: bool = False,
    force: bool = False,
    verbose: bool = False,
    debugging_mode: bool = False,
    lazy: bool = True,
    single_netcdf: bool = True,
) -> None:
    """Wrapper to run reader functions.

    Parameters
    ----------
    data_source : str
        Institution name (when campaign data spans more than 1 country) or country (when all campaigns (or sensor networks) are inside a given country)
    reader_name : str
        Campaign name
    raw_dir : str
        The directory path where all the raw content of a specific campaign is stored.
        The path must have the following structure:
            <...>/DISDRODB/Raw/<data_source>/<campaign_name>'.
        Inside the raw_dir directory, it is required to adopt the following structure:
        - /data/<station_id>/<raw_files>
        - /metadata/<station_id>.yaml
        Important points:
        - For each <station_id> there must be a corresponding YAML file in the metadata subfolder.
        - The <campaign_name> must semantically match between:
           - the raw_dir and processed_dir directory paths;
           - with the key 'campaign_name' within the metadata YAML files.
        - The campaign_name are expected to be UPPER CASE.
    processed_dir : str
        The desired directory path for the processed DISDRODB L0A and L0B products.
        The path should have the following structure:
            <...>/DISDRODB/Processed/<data_source>/<campaign_name>'
        For testing purpose, this function exceptionally accept also a directory path simply ending
        with <campaign_name> (i.e. /tmp/<campaign_name>).
    l0a_processing : bool
      Whether to launch processing to generate L0A Apache Parquet file(s) from raw data.
      The default is True.
    l0b_processing : bool
      Whether to launch processing to generate L0B netCDF4 file(s) from L0A data.
      The default is True.
    keep_l0a : bool
        Whether to keep the L0A files after having generated the L0B netCDF products.
        The default is False.
    force : bool
        If True, overwrite existing data into destination directories.
        If False, raise an error if there are already data into destination directories.
        The default is False.
    verbose : bool
        Whether to print detailed processing information into terminal.
        The default is False.
    debugging_mode : bool
        If True, it reduces the amount of data to process.
        - For L0A processing, it processes just 3 raw data files.
        - For L0B processing, it takes a small subset of the L0A Apache Parquet dataframe.
        The default is False.
    lazy : bool
        Whether to perform processing lazily with dask.
        If lazy=True, it employed dask.array and dask.dataframe.
        If lazy=False, it employed pandas.DataFrame and numpy.array.
        The default is True.
    single_netcdf : bool
        Whether to concatenate all raw files into a single L0B netCDF file.
        If single_netcdf=True, all raw files will be saved into a single L0B netCDF file.
        If single_netcdf=False, each raw file will be converted into the corresponding L0B netCDF file.
        The default is True.
    """

    # Get the corresponding reader function.
    reader = get_reader(data_source, reader_name)

    # Run the reader
    reader(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        l0a_processing=l0a_processing,
        l0b_processing=l0b_processing,
        keep_l0a=keep_l0a,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=lazy,
        single_netcdf=single_netcdf,
    )


def is_documented_by(original):
    """Wrapper function to apply generic docstring to the decorated function

    Parameters
    ----------
    original : function
        funtion to take the docstring from
    """

    def wrapper(target):
        target.__doc__ = original.__doc__
        return target

    return wrapper


def reader_generic_docstring():
    """Script to process raw data to L0A and L0B format.

    Parameters
    ----------
    raw_dir : str
        The directory path where all the raw content of a specific campaign is stored.
        The path must have the following structure:
            <...>/DISDRODB/Raw/<data_source>/<campaign_name>'.
        Inside the raw_dir directory, it is required to adopt the following structure:
        - /data/<station_id>/<raw_files>
        - /metadata/<station_id>.yaml
        Important points:
        - For each <station_id> there must be a corresponding YAML file in the metadata subfolder.
        - The <campaign_name> must semantically match between:
           - the raw_dir and processed_dir directory paths;
           - with the key 'campaign_name' within the metadata YAML files.
        - The campaign_name are expected to be UPPER CASE.
    processed_dir : str
        The desired directory path for the processed DISDRODB L0A and L0B products.
        The path should have the following structure:
            <...>/DISDRODB/Processed/<data_source>/<campaign_name>'
        For testing purpose, this function exceptionally accept also a directory path simply ending
        with <campaign_name> (i.e. /tmp/<campaign_name>).
    l0a_processing : bool
      Whether to launch processing to generate L0A Apache Parquet file(s) from raw data.
      The default is True.
    l0b_processing : bool
      Whether to launch processing to generate L0B netCDF4 file(s) from L0A data.
      The default is True.
    keep_l0a : bool
        Whether to keep the L0A files after having generated the L0B netCDF products.
        The default is False.
    force : bool
        If True, overwrite existing data into destination directories.
        If False, raise an error if there are already data into destination directories.
        The default is False.
    verbose : bool
        Whether to print detailed processing information into terminal.
        The default is False.
    debugging_mode : bool
        If True, it reduces the amount of data to process.
        - For L0A processing, it processes just 3 raw data files.
        - For L0B processing, it takes a small subset of the L0A Apache Parquet dataframe.
        The default is False.
    lazy : bool
        Whether to perform processing lazily with dask.
        If lazy=True, it employed dask.array and dask.dataframe.
        If lazy=False, it employed pandas.DataFrame and numpy.array.
        The default is True.
    single_netcdf : bool
        Whether to concatenate all raw files into a single L0B netCDF file.
        If single_netcdf=True, all raw files will be saved into a single L0B netCDF file.
        If single_netcdf=False, each raw file will be converted into the corresponding L0B netCDF file.
        The default is True.

    """



from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = "v0.0.12"
DESCRIPTION = "This package provides tools to homogenize, process, and analyze global disdrometer data."


# Setting up
setup(
    name="disdrodb",
    version=VERSION,
    author="Gionata Ghiggi",
    author_email="",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    # Always use forward '/', even on Windows
    # See https://setuptools.readthedocs.io/en/latest/userguide/datafiles.html#data-files-support
    package_data={
        "disdrodb": [
            "L0/configs/*/*.yml",
            "L0/readers/*/*.py",
        ]
    },
    install_requires=[
        "click",
        "h5py",
        "netCDF4",
        "pyarrow",
        "PyYAML",
        "setuptools",
        "xarray",
        "dask",
    ],
    keywords=["python", "disdrometer"],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    entry_points={
        "console_scripts": [
            "run_disdrodb_l0_reader=disdrodb.L0.run_disdrodb_l0_reader:run_reader_cmd",
        ]
    },
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 13 04:57:12 2022

@author: kimbo
"""
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
from disdrodb.L0 import run_L0


from disdrodb.L0.L0_processing import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    l0a_processing=True,
    l0b_processing=True,
    keep_l0a=False,
    force=False,
    verbose=False,
    debugging_mode=False,
    lazy=True,
    single_netcdf=True,
):

    ####----------------------------------------------------------------------.
    ###########################
    #### CUSTOMIZABLE CODE ####
    ###########################
    #### - Define raw data headers
    # Notes
    # - In all files, the datalogger voltage hasn't the delimeter,
    #   so need to be split to obtain datalogger_voltage and rainfall_rate_32bit

    column_names_temp = ["temp"]

    column_names = [
        "start_identifier",
        "device_address",
        "sensor_serial_number",
        "date_sensor",
        "time_sensor",
        "weather_code_synop_4677_5min",
        "weather_code_synop_4680_5min",
        "weather_code_metar_4678_5min",
        "precipitation_rate_5min",
        "weather_code_synop_4677",
        "weather_code_synop_4680",
        "weather_code_metar_4678",
        "precipitation_rate",
        "rainfall_rate",
        "snowfall_rate",
        "precipitation_accumulated",
        "mor_visibility",
        "reflectivity",
        "quality_index",
        "max_hail_diameter",
        "laser_status",
        "static_signal",
        "laser_temperature_analog_status",
        "laser_temperature_digital_status",
        "laser_current_analog_status",
        "laser_current_digital_status",
        "sensor_voltage_supply_status",
        "current_heating_pane_transmitter_head_status",
        "current_heating_pane_receiver_head_status",
        "temperature_sensor_status",
        "current_heating_voltage_supply_status",
        "current_heating_house_status",
        "current_heating_heads_status",
        "current_heating_carriers_status",
        "control_output_laser_power_status",
        "reserve_status",
        "temperature_interior",
        "laser_temperature",
        "laser_current_average",
        "control_voltage",
        "optical_control_voltage_output",
        "sensor_voltage_supply",
        "current_heating_pane_transmitter_head",
        "current_heating_pane_receiver_head",
        "temperature_ambient",
        "current_heating_voltage_supply",
        "current_heating_house",
        "current_heating_heads",
        "current_heating_carriers",
        "number_particles",
        "number_particles_internal_data",
        "number_particles_min_speed",
        "number_particles_min_speed_internal_data",
        "number_particles_max_speed",
        "number_particles_max_speed_internal_data",
        "number_particles_min_diameter",
        "number_particles_min_diameter_internal_data",
        "number_particles_no_hydrometeor",
        "number_particles_no_hydrometeor_internal_data",
        "number_particles_unknown_classification",
        "number_particles_unknown_classification_internal_data",
        "number_particles_class_1",
        "number_particles_class_1_internal_data",
        "number_particles_class_2",
        "number_particles_class_2_internal_data",
        "number_particles_class_3",
        "number_particles_class_3_internal_data",
        "number_particles_class_4",
        "number_particles_class_4_internal_data",
        "number_particles_class_5",
        "number_particles_class_5_internal_data",
        "number_particles_class_6",
        "number_particles_class_6_internal_data",
        "number_particles_class_7",
        "number_particles_class_7_internal_data",
        "number_particles_class_8",
        "number_particles_class_8_internal_data",
        "number_particles_class_9",
        "number_particles_class_9_internal_data",
        "raw_drop_number",
    ]

    ##------------------------------------------------------------------------.
    #### - Define reader options

    reader_kwargs = {}

    # - Define delimiter
    reader_kwargs["delimiter"] = "#"

    # - Avoid first column to become df index !!!
    reader_kwargs["index_col"] = False

    # - Define behaviour when encountering bad lines
    reader_kwargs["on_bad_lines"] = "skip"

    # - Define reader engine
    #   - C engine is faster
    #   - Python engine is more feature-complete
    reader_kwargs["engine"] = "python"

    # - Define on-the-fly decompression of on-disk data
    #   - Available: gzip, bz2, zip
    reader_kwargs["compression"] = "infer"

    # - Strings to recognize as NA/NaN and replace with standard NA flags
    #   - Already included: ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’,
    #                       ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’,
    #                       ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
    reader_kwargs["na_values"] = ["na", "", "error", "NA", "NP   "]

    # - Define max size of dask dataframe chunks (if lazy=True)
    #   - If None: use a single block for each file
    #   - Otherwise: "<max_file_size>MB" by which to cut up larger files
    reader_kwargs["blocksize"] = None  # "50MB"

    # Cast all to string
    reader_kwargs["dtype"] = str

    # Skip first row as columns names
    reader_kwargs["header"] = None

    # Skip first 3 rows
    reader_kwargs["skiprows"] = 3

    ##------------------------------------------------------------------------.
    #### - Define facultative dataframe sanitizer function for L0 processing
    # - Enable to deal with bad raw data files
    # - Enable to standardize raw data files to L0 standards  (i.e. time to datetime)
    df_sanitizer_fun = None

    def df_sanitizer_fun(df, lazy=False):
        # Import dask or pandas
        if lazy:
            import dask.dataframe as dd
        else:
            import pandas as dd

        # Drop header's log rows
        df = df.loc[df["temp"].astype(str).str.len() > 30]

        # Split first 80 columns
        df = df["temp"].str.split(";", n=79, expand=True)

        df.columns = column_names

        # Clean raw_drop_number (ignore last 5 column)
        df["raw_drop_number"] = df["raw_drop_number"].str[:1760]

        # Drop row if start_identifier different than 00
        df = df.loc[df["start_identifier"].astype(str) == "00"]

        # Merge time
        df["time"] = df["date_sensor"].astype(str) + " " + df["time_sensor"]
        # Drop rows with invalid date
        df = df.loc[df["time"].astype(str).str.len() == 17]
        try:
            df["time"] = dd.to_datetime(df["time"], format="%d.%m.%y %H:%M:%S")
        except ValueError:
            df["time"] = dd.to_datetime(
                df["time"], format="%d.%m.%y %H:%M:%S", errors="coerce"
            )
            df = df.loc[df.time.notnull()]

        # Drop invalid rows
        df = df.loc[df["raw_drop_number"].astype(str).str.len() == 1760]

        # Columns to drop
        columns_to_drop = [
            "start_identifier",
            "device_address",
            "sensor_serial_number",
            "date_sensor",
            "time_sensor",
            "weather_code_synop_4677_5min",
            "weather_code_synop_4680_5min",
            "weather_code_metar_4678_5min",
            "weather_code_metar_4678",
            "quality_index",
            "max_hail_diameter",
            "laser_status",
            "static_signal",
            "laser_temperature_analog_status",
            "laser_temperature_digital_status",
            "laser_current_analog_status",
            "laser_current_digital_status",
            "sensor_voltage_supply_status",
            "current_heating_pane_transmitter_head_status",
            "current_heating_pane_receiver_head_status",
            "temperature_sensor_status",
            "current_heating_voltage_supply_status",
            "current_heating_house_status",
            "current_heating_heads_status",
            "current_heating_carriers_status",
            "control_output_laser_power_status",
            "reserve_status",
            "temperature_interior",
            "laser_current_average",
            "optical_control_voltage_output",
            "sensor_voltage_supply",
            "current_heating_pane_transmitter_head",
            "current_heating_pane_receiver_head",
            "current_heating_voltage_supply",
            "current_heating_house",
            "current_heating_heads",
            "current_heating_carriers",
            "number_particles",
            "number_particles_internal_data",
            "number_particles_min_speed",
            "number_particles_min_speed_internal_data",
            "number_particles_max_speed",
            "number_particles_max_speed_internal_data",
            "number_particles_min_diameter",
            "number_particles_min_diameter_internal_data",
            "number_particles_no_hydrometeor",
            "number_particles_no_hydrometeor_internal_data",
            "number_particles_unknown_classification",
            "number_particles_unknown_classification_internal_data",
            "number_particles_class_1",
            "number_particles_class_1_internal_data",
            "number_particles_class_2",
            "number_particles_class_2_internal_data",
            "number_particles_class_3",
            "number_particles_class_3_internal_data",
            "number_particles_class_4",
            "number_particles_class_4_internal_data",
            "number_particles_class_5",
            "number_particles_class_5_internal_data",
            "number_particles_class_6",
            "number_particles_class_6_internal_data",
            "number_particles_class_7",
            "number_particles_class_7_internal_data",
            "number_particles_class_8",
            "number_particles_class_8_internal_data",
            "number_particles_class_9",
            "number_particles_class_9_internal_data",
        ]

        df = df.drop(columns=columns_to_drop)

        # Some values could be nan
        df = df.dropna()

        return df

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files in raw_dir/data/<station_id>
    files_glob_pattern = "*.txt*"

    ####----------------------------------------------------------------------.
    #### - Create L0 products
    run_L0(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        l0a_processing=l0a_processing,
        l0b_processing=l0b_processing,
        keep_l0a=keep_l0a,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=lazy,
        single_netcdf=single_netcdf,
        # Custom arguments of the reader
        files_glob_pattern=files_glob_pattern,
        column_names=column_names,
        reader_kwargs=reader_kwargs,
        df_sanitizer_fun=df_sanitizer_fun,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr  6 10:03:26 2022

@author: kimbo
"""

import os
import time

# Directory
from disdrodb.L0.io import check_directories
from disdrodb.L0.io import get_campaign_name
from disdrodb.L0.io import create_directory_structure

# Logger
from disdrodb.utils.logger import create_L0_logger
from disdrodb.utils.logger import close_logger

# Metadata
from disdrodb.L0.metadata import read_metadata
from disdrodb.L0.check_standards import check_sensor_name

# L0 processing
from disdrodb.L0.L0A_processing import get_file_list
from disdrodb.L0.io import get_L0B_fpath
from disdrodb.L0.L0B_processing import write_L0B
from disdrodb.L0.L0_processing import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    l0a_processing,
    l0b_processing,
    keep_l0a,
    force=True,
    verbose=True,
    debugging_mode=False,
    lazy=True,
    single_netcdf=False,
):

    # Define functions to reformat ARM netCDFs
    def reformat_ARM_files(file_list, attrs):
        """
        Reformat OTT Parsivel ARM netCDF files.

        Parameters
        ----------
        file_list : list
            Filepaths of NetCDFs to combine and reformat.
        attrs : dict
            DISDRODB metadata about the station.
        """
        from disdrodb.L0.utils_nc import xr_concat_datasets
        from disdrodb.L0.L0B_processing import get_coords
        from disdrodb.L0.auxiliary import get_ARM_LPM_dict
        from disdrodb.L0.standards import set_DISDRODB_L0_attrs

        sensor_name = attrs["sensor_name"]
        # --------------------------------------------------------
        #### Open netCDFs
        file_list = sorted(file_list)
        try:
            ds = xr_concat_datasets(file_list)
        except Exception as e:
            msg = f"Error in concatenating netCDF datasets. The error is: \n {e}"
            raise RuntimeError(msg)

        # --------------------------------------------------------
        # Select DISDRODB variable and rename
        dict_ARM = get_ARM_LPM_dict(sensor_name=sensor_name)
        vars_ARM = set(dict_ARM.keys())
        vars_ds = set(ds.data_vars)
        vars_selection = vars_ds.intersection(vars_ARM)
        dict_ARM_selection = {k: dict_ARM[k] for k in vars_selection}
        ds = ds[vars_selection]
        ds = ds.rename(dict_ARM_selection)

        # --------------------------------------------------------
        # Rename dimensions
        dict_dims = {
            "particle_size": "diameter_bin_center",
            "raw_fall_velocity": "velocity_bin_center",
        }
        ds = ds.rename_dims(dict_dims)
        ds = ds.rename(dict_dims)
        # --------------------------------------------------------
        # Update coordinates
        coords = get_coords(attrs["sensor_name"])
        coords["crs"] = attrs["crs"]
        coords["longitude"] = attrs["longitude"]
        coords["latitude"] = attrs["latitude"]
        coords["altitude"] = attrs["altitude"]
        ds = ds.drop(["latitude", "longitude", "altitude"])
        ds = ds.assign_coords(coords)

        # --------------------------------------------------------
        # Set DISDRODB attributes
        ds = set_DISDRODB_L0_attrs(ds, attrs)

        # --------------------------------------------------------
        return ds

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files in raw_dir/data/<station_id>
    raw_data_glob_pattern = "*.cdf"

    ####-------------------------------------------------------------------.
    ####################
    #### FIXED CODE ####
    ####################
    # ---------------------------------------------------------------------.
    # Initial directory checks
    raw_dir, processed_dir = check_directories(raw_dir, processed_dir, force=force)

    # Retrieve campaign name
    campaign_name = get_campaign_name(raw_dir)

    # --------------------------------------------------------------------.
    # Define logging settings
    logger = create_L0_logger(processed_dir, campaign_name)

    # ---------------------------------------------------------------------.
    # Create directory structure
    create_directory_structure(raw_dir, processed_dir)

    # ---------------------------------------------------------------------.
    # Get station list
    list_stations_id = os.listdir(os.path.join(raw_dir, "data"))

    # ---------------------------------------------------------------------.
    #### Loop over each station_id directory and process the files
    # station_id = list_stations_id[0]
    for station_id in list_stations_id:
        # ---------------------------------------------------------------------.
        t_i = time.time()
        msg = f" - Processing of station_id {station_id} has started"
        if verbose:
            print(msg)
        logger.info(msg)
        # ---------------------------------------------------------------------.
        # Retrieve metadata
        attrs = read_metadata(raw_dir=raw_dir, station_id=station_id)

        # Retrieve sensor name
        sensor_name = attrs["sensor_name"]
        check_sensor_name(sensor_name)

        # Retrieve list of files to process
        glob_pattern = os.path.join("data", station_id, raw_data_glob_pattern)
        file_list = get_file_list(
            raw_dir=raw_dir,
            glob_pattern=glob_pattern,
            verbose=verbose,
            debugging_mode=debugging_mode,
        )

        # -----------------------------------------------------------------.
        #### - Reformat netCDF to DISDRODB standards
        ds = reformat_ARM_files(file_list=file_list, attrs=attrs)

        # -----------------------------------------------------------------.
        #### - Save to DISDRODB netCDF standard
        fpath = get_L0B_fpath(processed_dir, station_id)
        write_L0B(ds, fpath=fpath, sensor_name=sensor_name)

        # -----------------------------------------------------------------.
        # End L0 processing
        t_f = time.time() - t_i
        msg = " - NetCDF standardization of station_id {} ended in {:.2f}s".format(
            station_id, t_f
        )
        if verbose:
            print(msg)
        logger.info(msg)
        msg = " --------------------------------------------------"
        if verbose:
            print(msg)
        logger.info(msg)
        # -----------------------------------------------------------------.
    # -----------------------------------------------------------------.
    msg = "### Script finish ###"
    print("\n  " + msg + "\n")
    logger.info(msg)

    close_logger(logger)
    # -----------------------------------------------------------------.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 23 18:44:52 2022

@author: ghiggi
"""
import os
import yaml
import logging

logger = logging.getLogger(__name__)


def _write_issue_timestamps_docs(f):
    """Provide template for timestamps.yml"""
    # TODO: Kim adapt to desired format
    f.write(
        "# This file is used to store dates to drop by the reader, the time format used is the isoformat (YYYY-mm-dd HH:MM:SS). \n"
    )
    f.write("# timestamp: list of timestamps \n")
    f.write("# time_period: list of list ranges of dates \n")
    f.write("# Example: \n")
    f.write(
        "# timestamp: ['2018-12-07 14:15','2018-12-07 14:17','2018-12-07 14:19', '2018-12-07 14:25'] \n"
    )
    f.write("# time_period: [['2018-08-01 12:00:00', '2018-08-01 14:00:00'], \n")
    f.write("#               ['2018-08-01 15:44:30', '2018-08-01 15:59:31'], \n")
    f.write("#               ['2018-08-02 12:44:30', '2018-08-02 12:59:31']] \n")


def create_issue_yml(
    fpath: str, timestamp: str = None, time_period: str = None
) -> None:
    """Create issue YAML file.

    Parameters
    ----------
    fpath : str
        timestamps.yml file path.
    timestamp : str, optional
        Timestamp, by default None
    time_period : str, optional
        Timeperiod, by default None
    """

    logger.info(f"Creating issue YAML file at {fpath}")
    with open(fpath, "w") as f:
        # Write template for timestamps.yml
        _write_issue_timestamps_docs(f)

        # Write timestamp if provided
        # TODO: Kim adapt to desired format
        if timestamp is not None:
            timestamp_dict = {"timestamp": timestamp}
            yaml.dump(timestamp_dict, f, default_flow_style=None)

        # Write timestamp if provided
        # TODO: Kim adapt to desired format
        if time_period is not None:
            time_period_dict = {"time_period": time_period}
            yaml.dump(time_period_dict, f, default_flow_style=None)


def read_issue(raw_dir: str, station_id: str) -> dict:
    """Read YAML issue file.

    Parameters
    ----------
    processed_dir : str
        Path of the processed directory
    station_id : int
        Id of the station.

    Returns
    -------
    dict
        Issue dictionary.
    """

    issue_fpath = os.path.join(raw_dir, "issue", station_id + ".yml")
    with open(issue_fpath, "r") as f:
        issue_dict = yaml.safe_load(f)
    # issue_dict = check_issue_compliance(issue_dict)
    return issue_dict


def check_issue_compliance(fpath: str) -> None:
    """Check issue compliance

    Parameters
    ----------
    fpath : str
        File pazh


    """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 18 12:20:46 2022

@author: ghiggi
"""

# TODO: adapt to new readers !!!!


def get_reader_cmd(
    reader_filepath,
    raw_dir,
    processed_dir,
    l0a_processing=True,
    l0b_processing=True,
    keep_l0a=False,
    force=False,
    verbose=False,
    debugging_mode=False,
    lazy=True,
    single_netcdf=True,
):
    """Create command to launch reader processing from Terminal."""
    # reader_TICINO_2018.py [OPTIONS] <raw_dir> <processed_dir>
    cmd = "".join(
        [
            "python",
            " ",
            reader_filepath,
            " ",
            "--l0a_processing=",
            str(l0a_processing),
            " ",
            "--l0b_processing=",
            str(l0b_processing),
            " ",
            "--force=",
            str(force),
            " ",
            "--verbose=",
            str(verbose),
            " ",
            "--debugging_mode=",
            str(debugging_mode),
            " ",
            "--lazy=",
            str(lazy),
            " ",
            "--keep_l0a=",
            str(keep_l0a),
            " ",
            "--single_netcdf=",
            str(single_netcdf),
            " ",
            raw_dir,
            " ",
            processed_dir,
        ]
    )
    return cmd



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

from disdrodb.L0 import run_L0


from disdrodb.L0.L0_processing import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    l0a_processing=True,
    l0b_processing=True,
    keep_l0a=False,
    force=False,
    verbose=False,
    debugging_mode=False,
    lazy=True,
    single_netcdf=True,
):

    ####----------------------------------------------------------------------.
    ###########################
    #### CUSTOMIZABLE CODE ####
    ###########################
    #### - Define raw data headers
    column_names = ["TEMPORARY"]

    ##------------------------------------------------------------------------.
    #### - Define reader options
    reader_kwargs = {}
    # - Define delimiter
    reader_kwargs["delimiter"] = "\\n"

    # - Avoid first column to become df index !!!
    reader_kwargs["index_col"] = False

    # - Define behaviour when encountering bad lines
    reader_kwargs["on_bad_lines"] = "skip"

    # Skip the first row (header)
    reader_kwargs["skiprows"] = 0

    # - Define reader engine
    #   - C engine is faster
    #   - Python engine is more feature-complete
    reader_kwargs["engine"] = "python"

    # - Define on-the-fly decompression of on-disk data
    #   - Available: gzip, bz2, zip
    reader_kwargs["compression"] = "infer"

    # - Strings to recognize as NA/NaN and replace with standard NA flags
    #   - Already included: ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’,
    #                       ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’,
    #                       ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
    reader_kwargs["na_values"] = ["na", "", "error"]

    # - Define max size of dask dataframe chunks (if lazy=True)
    #   - If None: use a single block for each file
    #   - Otherwise: "<max_file_size>MB" by which to cut up larger files
    reader_kwargs["blocksize"] = None  # "50MB"

    # Skip first row as columns names
    reader_kwargs["header"] = None

    # - Define encoding
    reader_kwargs["encoding"] = "ISO-8859-1"

    ##------------------------------------------------------------------------.
    #### - Define facultative dataframe sanitizer function for L0 processing
    # - Enable to deal with bad raw data files
    # - Enable to standardize raw data files to L0 standards  (i.e. time to datetime)
    def df_sanitizer_fun(df, lazy=False):
        import numpy as np

        # Import dask or pandas
        if lazy:
            import pandas as dd

            df = df.compute()
        else:
            import pandas as dd

        # Reshape dataframe and define dummy column names
        # - Assume always 97 fields
        n_fields = 97
        arr = df.to_numpy()
        arr = arr.reshape(int(len(arr) / n_fields), n_fields)
        df = dd.DataFrame(arr)
        df.columns = np.arange(1, n_fields + 1)

        # Define known field names
        column_names_dict = {
            1: "rainfall_rate_32bit",
            2: "rainfall_accumulated_32bit",
            3: "weather_code_synop_4680",
            4: "weather_code_synop_4677",
            5: "weather_code_metar_4678",
            6: "weather_code_nws",
            7: "reflectivity_32bit",
            8: "mor_visibility",
            9: "sample_interval",
            10: "laser_amplitude",
            11: "number_particles",
            12: "sensor_temperature",
            13: "sensor_serial_number",
            14: "firmware_iop",
            15: "firmware_dsp",
            16: "sensor_heating_current",
            17: "sensor_battery_voltage",
            18: "sensor_status",
            19: "start_time",
            20: "sensor_time",
            21: "sensor_date",
            22: "station_name",
            23: "station_number",
            24: "rainfall_amount_absolute_32bit",
            25: "error_code",
            # 26: 'sensor_temperature_pcb',
            # 28: 'sensor_temperature_receiver',
            # 29: 'sensor_temperature_trasmitter',
            30: "rainfall_rate_16_bit_30",
            31: "rainfall_rate_16_bit_1200",
            32: "rainfall_accumulated_16bit",
            # 33: 'reflectivity_16bit',
            # 34: 'rain_kinetic_energy',
            # 35: 'snowfall_rate',
            # 60: 'number_particles_all',
            # 61: 'list_particles',
            90: "raw_drop_concentration",
            91: "raw_drop_average_velocity",
            92: "raw_drop_number",
        }  # this correspond to the field '93: <value>' in the file

        # Rename columns and drop unknown field names
        column_names = list(column_names_dict.values())
        df = df.rename(column_names_dict, axis=1)
        df = df[column_names]

        # Drop columns not meeting DISDRODB standard
        col_to_drop = [
            "sensor_serial_number",
            "firmware_iop",
            "firmware_dsp",
            "station_name",
            "station_number",
            "sample_interval",
        ]
        df = df.drop(columns=col_to_drop)

        # Remove field number. Example: '<field_number>: 24.8' becomes '24.8'
        for col in df:
            df[col] = df[col].str[3:]

        # Drop columns where values have been not logged
        # - This varies between stations
        # - We assume that the logging does not change across files !!!
        df = df.replace("", np.nan)
        df = df.dropna(how="all", axis=1)

        # Define time
        df["time"] = dd.to_datetime(
            df["sensor_date"] + "-" + df["sensor_time"], format="%d.%m.%Y-%H:%M:%S"
        )
        df = df.drop(columns=["sensor_date", "sensor_time", "start_time"])

        # Reformat weather codes
        if "weather_code_metar_4678" in df.columns:
            df["weather_code_metar_4678"] = df["weather_code_metar_4678"].str.strip()
        if "weather_code_nws" in df.columns:
            df["weather_code_nws"] = df["weather_code_nws"].str.strip()

        return df

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files in raw_dir/data/<station_id>
    files_glob_pattern = "*.dat"  # There is only one file without extension

    ####----------------------------------------------------------------------.
    #### - Create L0 products
    run_L0(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        l0a_processing=l0a_processing,
        l0b_processing=l0b_processing,
        keep_l0a=keep_l0a,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=False,
        single_netcdf=single_netcdf,
        # Custom arguments of the reader
        files_glob_pattern=files_glob_pattern,
        column_names=column_names,
        reader_kwargs=reader_kwargs,
        df_sanitizer_fun=df_sanitizer_fun,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan 11 13:00:07 2022

@author: kimbo
"""
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

from disdrodb.L0 import run_L0


from disdrodb.L0.L0_processing import reader_generic_docstring, is_documented_by


@is_documented_by(reader_generic_docstring)
def reader(
    raw_dir,
    processed_dir,
    l0a_processing=True,
    l0b_processing=True,
    keep_l0a=False,
    force=False,
    verbose=False,
    debugging_mode=False,
    lazy=True,
    single_netcdf=True,
):

    ####----------------------------------------------------------------------.
    ###########################
    #### CUSTOMIZABLE CODE ####
    ###########################
    #### - Define raw data headers
    # Notes
    # - In all files, the datalogger voltage hasn't the delimeter,
    #   so need to be split to obtain datalogger_voltage and rainfall_rate_32bit

    column_names = [
        "temp1",
        "temp2",
        "temp3",
        "temp4",
        "temp5",
        "temp6",
        "temp7",
        "temp8",
        "temp9",
        "temp10",
        "temp11",
        "temp12",
        "temp13",
        "temp14",
        "temp15",
        "temp16",
        "temp17",
        "temp18",
        "temp19",
        "temp20",
        "temp21",
        "temp22",
        "temp23",
    ]

    ##------------------------------------------------------------------------.
    #### - Define reader options
    reader_kwargs = {}
    # - Define delimiter
    reader_kwargs["delimiter"] = ";"

    # - Avoid first column to become df index !!!
    reader_kwargs["index_col"] = False

    # - Define behaviour when encountering bad lines
    reader_kwargs["on_bad_lines"] = "skip"

    # - Define reader engine
    #   - C engine is faster
    #   - Python engine is more feature-complete
    reader_kwargs["engine"] = "python"

    # - Define on-the-fly decompression of on-disk data
    #   - Available: gzip, bz2, zip
    reader_kwargs["compression"] = "infer"

    # - Strings to recognize as NA/NaN and replace with standard NA flags
    #   - Already included: ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’, ‘-1.#QNAN’,
    #                       ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’, ‘<NA>’, ‘N/A’,
    #                       ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’
    reader_kwargs["na_values"] = [
        "na",
        "",
        "error",
        "NA",
        "Error in data reading! 0000.000",
    ]

    # - Define max size of dask dataframe chunks (if lazy=True)
    #   - If None: use a single block for each file
    #   - Otherwise: "<max_file_size>MB" by which to cut up larger files
    reader_kwargs["blocksize"] = None  # "50MB"

    # Cast all to string
    reader_kwargs["dtype"] = str

    # Different enconding for this campaign
    reader_kwargs["encoding"] = "latin-1"

    # Skip first row as columns names
    reader_kwargs["header"] = None

    ##------------------------------------------------------------------------.
    #### - Define facultative dataframe sanitizer function for L0 processing
    # - Enable to deal with bad raw data files
    # - Enable to standardize raw data files to L0 standards  (i.e. time to datetime)
    df_sanitizer_fun = None

    def df_sanitizer_fun(df, lazy=False):
        # Import dask or pandas
        if lazy:
            import dask.dataframe as dd
        else:
            import pandas as dd

        column_names = [
            "time",
            "latitude",
            "longitude",
            "weather_code_synop_4680",
            "weather_code_synop_4677",
            "reflectivity_32bit",
            "mor_visibility",
            "laser_amplitude",
            "number_particles",
            "sensor_temperature",
            "sensor_heating_current",
            "sensor_battery_voltage",
            "datalogger_error",
            "rainfall_amount_absolute_32bit",
            "All_0",
            "raw_drop_concentration",
            "raw_drop_average_velocity",
            "raw_drop_number",
        ]

        column_names_2 = [
            "id",
            "latitude",
            "longitude",
            "time",
            "all_nan",
            "rainfall_rate_32bit",
            "rainfall_accumulated_32bit",
            "weather_code_synop_4680",
            "weather_code_synop_4677",
            "reflectivity_32bit",
            "mor_visibility",
            "laser_amplitude",
            "number_particles",
            "sensor_temperature",
            "sensor_heating_current",
            "sensor_battery_voltage",
            "All_0",
            "rainfall_amount_absolute_32bit",
            "datalogger_error",
            "raw_drop_concentration",
            "raw_drop_average_velocity",
            "raw_drop_number",
            "End_line",
        ]

        # - Drop all nan in latitude (define in reader_kwargs['na_values'])
        df = df[~df.iloc[:, 1].isna()]
        if len(df.index) == 0:
            df.columns = [
                "latitude",
                "longitude",
                "time",
                "rainfall_rate_32bit",
                "rainfall_accumulated_32bit",
                "weather_code_synop_4680",
                "weather_code_synop_4677",
                "reflectivity_32bit",
                "mor_visibility",
                "laser_amplitude",
                "number_particles",
                "sensor_temperature",
                "sensor_heating_current",
                "sensor_battery_voltage",
                "rainfall_amount_absolute_32bit",
                "raw_drop_concentration",
                "raw_drop_average_velocity",
                "raw_drop_number",
            ]
            # df = df.iloc[:,:18]
            return df

        # - If first column is ID, than is a different format

        if lazy:
            flag = df.iloc[:, 0].str.isnumeric().all().compute()
        else:
            flag = df.iloc[:, 0].str.isnumeric().all()

        if flag:
            # - Rename columns
            df.columns = column_names_2
            # - Remove ok from rainfall_rate_32bit
            if lazy:
                df["rainfall_rate_32bit"] = df["rainfall_rate_32bit"].str.replace(
                    "OK,", ""
                )
            else:
                # df['rainfall_rate_32bit'] = df['rainfall_rate_32bit'].str.split(',').str[-1]
                # - Suppress SettingWithCopyWarning error (A value is trying to be set on a copy of a slice from a DataFrame)
                dd.options.mode.chained_assignment = None
                df["rainfall_rate_32bit"] = (
                    df["rainfall_rate_32bit"].str.split(",").str[-1]
                )

            # - Drop useless columns
            col_to_drop = ["id", "all_nan", "All_0", "datalogger_error", "End_line"]
            df = df.drop(columns=col_to_drop)

            # - Check latutide and longitute
            df = df.loc[df["latitude"].astype(str).str.len() < 11]
            df = df.loc[df["longitude"].astype(str).str.len() < 11]

            # - Convert time column to datetime
            df["time"] = dd.to_datetime(df["time"], errors="coerce")
            df = df.dropna()
            if len(df.index) == 0:
                for col in col_to_drop:
                    column_names_2.remove(col)
                df.columns = column_names_2
                return df
            df["time"] = dd.to_datetime(df["time"], format="%d-%m-%Y %H:%M:%S")

        else:

            # - Drop excedeed columns
            df = df.iloc[:, :18]
            # - Rename columns
            df.columns = column_names
            # - Drop useless columns
            col_to_drop = ["All_0", "datalogger_error"]
            df = df.drop(columns=col_to_drop)
            # - Convert time column to datetime
            df["time"] = dd.to_datetime(df["time"], errors="coerce")
            df = df.dropna()
            if len(df.index) == 0:
                for col in col_to_drop:
                    column_names.remove(col)
                df.columns = column_names
                return df
            df["time"] = dd.to_datetime(df["time"], format="%d/%m/%Y %H:%M:%S")

        # Drop latitude, longitude
        df = df.drop(columns=["longitude", "latitude"])

        # - Drop columns if nan
        col_to_drop_if_na = [
            "raw_drop_concentration",
            "raw_drop_average_velocity",
            "raw_drop_number",
        ]
        df = df.dropna(subset=col_to_drop_if_na)

        # - Drop invalid raw_drop_concentration, raw_drop_average_velocity and raw_drop_number
        df = df.loc[df["raw_drop_concentration"].astype(str).str.len() == 224]
        df = df.loc[df["raw_drop_average_velocity"].astype(str).str.len() == 224]
        df = df.loc[df["raw_drop_number"].astype(str).str.len() == 4096]

        df = df[df["raw_drop_number"].str.contains("0\x100") == False]

        # # - Cast dataframe to dtypes
        # from disdrodb.data_encodings import get_L0_dtype_standards
        # dtype_dict = get_L0_dtype_standards(sensor_name=sensor_name)

        # dtype_dict_not_object = {}
        # for k, v in dtype_dict.items():
        #     if v != 'object':
        #         dtype_dict_not_object[k] =  v
        # dtype_dict_not_object.pop('time')

        # for column in df.columns:
        #     if column in dtype_dict_not_object:
        #         df[column] = dd.to_numeric(df[column], errors='coerce')
        #         invalid_rows_index = df.loc[df[column].isna()].index
        #         if lazy:
        #             if invalid_rows_index.size.compute() != 0:
        #                 df = df.dropna(subset=[column])
        #         else:
        #             if invalid_rows_index.size != 0:
        #                 df = df.dropna(subset=[column])
        #                 # df = df.drop(invalid_rows_index)
        #         df[column] = df[column].astype(dtype_dict[column])

        return df

    ##------------------------------------------------------------------------.
    #### - Define glob pattern to search data files in raw_dir/data/<station_id>
    files_glob_pattern = "*.log*"

    ####----------------------------------------------------------------------.
    #### - Create L0 products
    run_L0(
        raw_dir=raw_dir,
        processed_dir=processed_dir,
        l0a_processing=l0a_processing,
        l0b_processing=l0b_processing,
        keep_l0a=keep_l0a,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=lazy,
        single_netcdf=single_netcdf,
        # Custom arguments of the reader
        files_glob_pattern=files_glob_pattern,
        column_names=column_names,
        reader_kwargs=reader_kwargs,
        df_sanitizer_fun=df_sanitizer_fun,
    )
